#! /usr/bin/env python
#
# Provides class Main to encompass common script processing. By default, the
# command line arguments are analyzed to determine optional filename, which is
# opened. Then, the input stream is feed line-by-line into the process_line
# method.
#
# Usage exmple:
#    from main import Main
#
#    class MyMain(Main):
#       def process_line(self, line):
#           if "funny" in line:
#               print("Funny looking line: %s" % line)
#
#    if __name__ == "__main__":
#        MyMain().run()
#    
# Notes:
# - See simple_main_example.py for a non-trivial example.
# - To add command-line arguments, pass corresponding arguments to Main's
#   initialization. For example,
#      Main(boolean_options=["fubar"], 
#           int_options=[("count", "Number of times", 10)]).run()
# - As the class likely will just be instantiated once, initialization
#   can be simplfied by using class-level variables for options, as follows:
#      Script(Main):
#          count = 5
#          fubar = False
#          def setup(self):
#              fubar = self.get_parsed_option("fubar")
# - With non-trivial command processing (e.g., positional arguments), it 
#   might be better to do this in the constructor, as follows:
#       def __init__(*args, **kwargs):
#           super(MyMain, self).__init__(*args, positional_options=["targets"], 
#                                        **kwargs)
# - Changes to temporary directory/file support should be synchronized with the
#   unit testing base class (see tests/unittest_wrapper.py.
# - Overriding the temporary directory can be handy during debugging (via
#   TEMP_BASE or TEMP_FILE). If you invoke sub-scripts, you might need to
#   specify different ones, as in adhoc/optimize_company_extraction.py.
# - During page-tracking mode, the page numbers are set based on occurrence of
#   form feed characters: \f (n.b., same as ^L and 0x0c).
# - A form feed is treated as an implicit paragraph break: see read_input.
# - The input processing can be used in non-Main scripts by creating a
#   dummy instance and then calling read_input (see randomize_lines.py):
#      dummy = Main([]);   dummy.input_stream = str
#      for line in dummy.process_input(): ...
#
# Note:
# - PERL_SWITCH_PARSING allows for Perl-style -var=val command switches. This 
#   was added to facilitate porting Perl scripts, especially those used in aliases. See
#      https://github.com/tomasohara/shell-scripts/blob/main/tomohara-aliases.bash
#   For example, https://github.com/tomasohara/shell-scripts/blob/main/check_errors.py.
# - This requires a workaround due to an argparse limitation:
#      https://github.com/spotify/luigi/issues/193 [boolean as command-line arg]
#
# TODO:
# - *** Convert tpo_common calls (i.e., tpo.xyz) to debug!
# - * Clarify TEMP_BASE vs. TEMP_FILE usage.
# - ** Create cheatsheet for argparse tricks (e.,g., using argparse.SUPPRESS to hide).
# - Specify argument via input dicts, such as in 
#      options=[{"name": "fubar", "type": bool}, 
#               {"name": "count", type: int, default: 10}]
# - Add support for perl-style paragraph mode in input processing.
# - Add support for multple input files (e.g., via fileinput module).
# - Add support for csv.csv_reader (see usage in cut.py).
# - Add support for argument aliases (e.g., --input-delim for --delim).
# - Have option for processing text by page by page, instead of
#   just tracking page number with
# - Clarify the input processing in various modes: line, paragraph and file-input.
# - Add function for getting temp_base as dir:
#       dummy_app = Main([], use_temp_base_dir=True)
#       temp_wav_path = gh.form_path(dummy_app.temp_base, "sample.wav")
#

"""Module for encapsulating main() processing"""

# Standard packages
import argparse
import io
import os
import re
import sys
import tempfile

# Local packages
from mezcla import debug
from mezcla import tpo_common as tpo
from mezcla import glue_helpers as gh
from mezcla import system
from mezcla.my_regex import my_re
from mezcla.system import getenv_bool

# Constants
HELP_ARG = "--help"
USAGE_ARG = "--usage"
VERBOSE_ARG = "verbose"
USE_PARAGRAPH_MODE_DEFAULT = getenv_bool("USE_PARAGRAPH_MODE", False)
USE_PARAGRAPH_MODE = getenv_bool("PARAGRAPH_MODE", USE_PARAGRAPH_MODE_DEFAULT,
                                 "Process input in Perl-style paragraph mode")
FILE_INPUT_MODE = getenv_bool("FILE_INPUT_MODE", False,
                              "Read stdin using Perl-style file input mode--aka file slurping")
debug.assertion(not (USE_PARAGRAPH_MODE and FILE_INPUT_MODE))

FORM_FEED = "\f"
TRACK_PAGES = getenv_bool("TRACK_PAGES", False,
                          "Track page boundaries by form feed--\\f or ^L")
RETAIN_FORM_FEED = getenv_bool("RETAIN_FORM_FEED", False,
                               "Include formfeed (\\f) at start of each segment")
DEFAULT_FILE_BASE = my_re.sub(r".py\w*$", "", gh.basename(__file__))
FILE_BASE = system.getenv_text("FILE_BASE", DEFAULT_FILE_BASE,
                               "Basename for output files including dir")
SHOW_ENV_OPTIONS = system.getenv_bool("ENV_USAGE", debug.detailed_debugging(),
                                      "Include environment options in usage")
# TODO: Put following in common module
INDENT = system.getenv_text("INDENT", "    ",
                            "Indentation for system output")
BRIEF_USAGE = system.getenv_bool("BRIEF_USAGE", False,
                                 "Show brief usage with autohelp")
PERL_SWITCH_PARSING = system.getenv_bool("PERL_SWITCH_PARSING", False,
                                         "Preprocess args to expand Perl-style -var[=[val=1]] to --var=val")
## HACK: This is needed if boolean options default to true based on run-time initialization
NEGATIVE_BOOL_ARGS = system.getenv_bool("NEGATIVE_BOOL_ARGS", False,
                                        "Add negation option for each boolean option")
SHORT_OPTIONS = system.getenv_bool("SHORT_OPTIONS", False,
                                   "Automatically derive short options")
ENV_OPTION_PREFIX = system.getenv_value("ENV_OPTION_PREFIX", None,
                                        "Environment variable prefix to check for default")
## TEST
## TEMP_BASE = system.getenv_value("TEMP_BASE", None,
##                                 "Override for temporary file basename")
TEMP_BASE = gh.TEMP_BASE
## OLD:
## USE_TEMP_BASE_DIR = system.getenv_bool("USE_TEMP_BASE_DIR", False,
##                                        "Whether TEMP_BASE should be a dir instead of prefix")
USE_TEMP_BASE_DIR = gh.USE_TEMP_BASE_DIR
## TEST
## TEMP_FILE = system.getenv_value("TEMP_FILE", None,
##                                 "Override for temporary filename")
TEMP_FILE = gh.TEMP_FILE
KEEP_TEMP_FILES = system.getenv_bool("KEEP_TEMP_FILES", debug.detailed_debugging(),
                                     "Retain temporary files")

#-------------------------------------------------------------------------------

class Main(object):
    """Class encompassing common script processing"""
    argument_parser = None
    force_unicode = False
    # TODO: add more class-wide member
    ## temp_base, temp_file
    verbose = False

    def __init__(self, runtime_args=None, description=None, skip_args=False,
                 # TODO: Either rename xyz_optiom to match python type name 
                 # or rename them without abbreviations.
                 # TODO: explain difference between positional_options and positional_arguments
                 multiple_files=False,
                 use_temp_base_dir=None,
                 usage_notes=None,
                 program=None,
                 paragraph_mode=None, track_pages=None, file_input_mode=None, newlines=None,
                 boolean_options=None, text_options=None, int_options=None,
                 float_options=None, positional_options=None, positional_arguments=None,
                 skip_input=None, manual_input=None, auto_help=None, brief_usage=None,
                 short_options=None, **kwargs):
        """Class constructor: parses RUNTIME_ARGS (or command line), with specifications
        for BOOLEAN_OPTIONS, TEXT_OPTIONS, INT_OPTIONS, FLOAT_OPTIONS, and POSITIONAL_OPTIONS
        (see convert_option). Includes options to SKIP_INPUT, or to have MANUAL_INPUT, or to use AUTO_HELP invocation (i.e., assuming {ha} if no args). Also allows for SHORT_OPTIONS"""
        tpo.debug_format("Main.__init__({args}, d={desc}, b={bools}, t={texts},"
                         + " i={ints}, f={floats}, po={posns}, pa={pargs}, si={skip}, m={mi}, ah={auto}, bu={usage},"
                         + " pm={para}, tp={page}, fim={file}, nl={nl}, prog={prog}, sa={skip_args},"
                         + " mult={mf}, temp_base={utbd} notes={us} short={so} kw={kw})", 5,
                         args=runtime_args, desc=description, bools=boolean_options,
                         texts=text_options, ints=int_options, floats=float_options,
                         mf=multiple_files, utbd=use_temp_base_dir,
                         posns=positional_options, pargs=positional_arguments, skip=skip_input, mi=manual_input,
                         auto=auto_help, usage=brief_usage, us=usage_notes,
                         para=paragraph_mode, page=track_pages, file=file_input_mode, nl=newlines,
                         prog=program, ha=HELP_ARG, skip_args=skip_args, so=short_options, kw=kwargs)
        self.description = "TODO: what the script does"   # *** DONT'T MODIFY: default TODO note for client
        self.boolean_options = []
        self.text_options = []
        self.int_options = []
        self.float_options = []
        self.positional_options = []
        self.process_line_warning = False
        self.input_stream = None
        self.end_of_page = False
        # TODO: line_num => total_lines_seen AND rel_line_num => line_num
        # TODO: para_num => total_paras_seen AND rel_para_num => para_num
        self.rel_line_num = -1
        self.rel_para_num = -1
        self.page_num = -1
        self.para_num = -1
        self.line_num = -1
        self.char_offset = -1
        self.raw_line = None
        # Note: manual_input was introduced after skip_input to allow for input processing
        # in bulk (e.g., via read_input generator). By default, neither is specified
        # (see template.py), and both should be assumed false.
        # TODO: *** Add better sanity checking (such as a filename on command line).
        if manual_input is None:
            # NOTE: skip_input=>manual_input: T=>T  F=>F  None=>F (i.e., bool(skip_input))
            manual_input = False if (skip_input is None) else skip_input
            debug.trace_fmt(7, "inferred manual_input: {mi}", mi=manual_input)
        self.manual_input = manual_input
        if skip_input is None:
            skip_input = self.manual_input
            debug.trace_fmt(7, "inferred skip_input: {si}", si=skip_input)
        self.skip_input = skip_input
        #
        self.parser = None
        if brief_usage is None:
            brief_usage = BRIEF_USAGE
        self.brief_usage = brief_usage  # show brief usage instead of full --help
        if auto_help is None:
            ## TODO: rework to be default if none specified for both skip_input and manual_input
            ## OLD: auto_help = self.skip_input
            auto_help = self.skip_input or not self.manual_input
            debug.trace(7, f"inferred auto_help: {auto_help}")
        self.auto_help = auto_help      # adds --help to command line if no arguments
        if usage_notes is None:
            usage_notes = ""
        self.notes = usage_notes
        if paragraph_mode is None:
            paragraph_mode = USE_PARAGRAPH_MODE
        self.paragraph_mode = paragraph_mode
        if file_input_mode is None:
            file_input_mode = FILE_INPUT_MODE
        self.file_input_mode = file_input_mode
        self.newlines = newlines
        if track_pages is None:
            track_pages = TRACK_PAGES
        self.track_pages = track_pages
        self.short_options = (short_options if (short_options is not None) else SHORT_OPTIONS)

        # Check miscellaneous options
        BINARY_INPUT_OPTION = "binary_input"
        PERL_SWITCH_PARSING_OPTION = "perl_switch_parsing"
        bad_options = system.difference(kwargs.keys(), [BINARY_INPUT_OPTION, PERL_SWITCH_PARSING_OPTION])
        debug.assertion(not bad_options, f"Extraneous kwargs: {bad_options}")
        self.binary_input = kwargs.get(BINARY_INPUT_OPTION, False)

        # Setup temporary file and/or base directory
        # Note: Uses NamedTemporaryFile (hence ntf_args)
        # TODO: allow temp_base handling to be overridable by constructor options
        # TODO: reconcile with unittest_wrapper.py.get_temp_dir
        prefix = (FILE_BASE + "-")
        ntf_args = {"prefix": prefix,
                    "delete": not debug.detailed_debugging(),
                    ## TODO: "suffix": "-"
                    }
        self.temp_base = (TEMP_BASE or tempfile.NamedTemporaryFile(**ntf_args).name)
        # TODO: self.use_temp_base_dir = gh.dir_exists(gh.basename(self.temp_base))
        # -or-: temp_base_dir = system.getenv_text("TEMP_BASE_DIR", " "); self.use_temp_base_dir = bool(temp_base_dir.strip()); ...
        if use_temp_base_dir is None:
            use_temp_base_dir = USE_TEMP_BASE_DIR
        self.use_temp_base_dir = use_temp_base_dir
        if self.use_temp_base_dir:
            ## TODO: gh.full_mkdir
            ## TEMP HACK: remove file if not a dir (n.b., quirk with NamedTemporaryFile
            if system.is_regular_file(self.temp_base):
                gh.delete_file(self.temp_base)
            gh.run("mkdir -p {dir}", dir=self.temp_base)
            ## TODO3: main-temp.txt???
            default_temp_file = gh.form_path(self.temp_base, "temp.txt")
        else:
            default_temp_file = self.temp_base
        self.temp_file = (TEMP_FILE or default_temp_file)

        # Note: --help assumed for input-less scripts with command line options
        # to avoid inadvertent script processing.
        #
        if ((runtime_args is None) and (not skip_args)):
            runtime_args = sys.argv[1:]
            debug.trace(4, f"Using sys.argv[1:] for runtime args: {runtime_args}")
            if self.auto_help and not runtime_args:
                help_arg = (USAGE_ARG if self.brief_usage else HELP_ARG)
                debug.trace(4, f"FYI: Adding {help_arg} to command line (as per auto_help)")
                runtime_args = [help_arg]
        #
        # Process special hook for converting Perl-style switches like -fu=123 to --fu=123
        # See -s option under perlrun man page for enabling this rudimentary switch parsing.
        # Note: mainly just intended for when porting Perl scripts.
        self.perl_switch_parsing = kwargs.get(PERL_SWITCH_PARSING_OPTION, PERL_SWITCH_PARSING)
        if self.perl_switch_parsing and runtime_args:
            debug.trace(4, "FYI: Enabling Perl-style options")
            debug.assertion(not re.search(r"--\w+", " ".join(runtime_args)),
                            "Shouldn't use Python arguments with PERL_SWITCH_PARSING")
            for i, arg in enumerate(runtime_args):
                if arg in ["-", "--"]:
                    break
                if my_re.search(r"^-([a-z0-9_]+\w*)=?(.*)$", arg, flags=re.IGNORECASE):
                    option = my_re.group(1)
                    value = (my_re.group(2) if len(my_re.group(2)) else "1")
                    new_arg = f"--{option}={value}"
                    debug.trace(4, f"Converted Perl-style arg {i} from {arg!r} to {new_arg}")
                    debug.assertion(not system.file_exists(arg))
                    runtime_args[i] = new_arg
                    
        # Get other options
        self.program = program
        if description:
            self.description = description
        if boolean_options:
            self.boolean_options += boolean_options
        # note: adds --verbose unless already specified (TODO: add way to disable)
        boolean_options_proper = [t for t in self.boolean_options if isinstance(t, str)]
        boolean_options_proper += [t[0] for t in self.boolean_options if isinstance(t, (list, tuple))]
        if (VERBOSE_ARG not in boolean_options_proper):
            debug.trace(6, f"Adding {VERBOSE_ARG} to {self.boolean_options}")
            self.boolean_options += [(VERBOSE_ARG, "Verbose output mode")]
        if text_options:
            self.text_options = text_options
        if int_options:
            self.int_options = int_options
        if float_options:
            self.float_options = float_options
        if positional_options or positional_arguments:
            # TODO: mark positional_options as deprecated
            debug.assertion(not (positional_options and positional_arguments))
            self.positional_options = positional_options or positional_arguments
        self.multiple_files = multiple_files      # sets other_filenames if multiple w/ nargs=+ 
        # Set defaults
        self.parsed_args = None
        self.filename = None
        self.other_filenames = []
        # Do command-line parsing
        # TODO: consolidate with runtime_args check above
        if not skip_args:
            self.check_arguments(runtime_args)
        debug.trace_current_context(level=debug.QUITE_DETAILED)
        debug.trace_object(6, self, label="Main instance")
        debug.trace_fmt(tpo.QUITE_DETAILED, "end of Main.__init__(); self={s}",
                        s=self)
        return

    def get_arguments(self, just_positional=False, just_optional=False):
        """Return list of arguments, optional just positional"""
        argument_specs = []
        if not just_positional:
            argument_specs += self.boolean_options + self.text_options + self.int_options + self.float_options
        if not just_optional:
            argument_specs += self.positional_options
        arguments = [(spec[0] if list(spec[0]) else spec) for spec in argument_specs]
        debug.trace(6, f"get_arguments([pos?={just_positional}, opt?={just_optional}] => {arguments}")
        return arguments
    
    def convert_option(self, option_spec, default_value=None, positional=False):
        """Convert OPTION_SPEC to (label, description, default) tuple. 
        Notes: The description and default of the specification are optional,
        and the parentheses can be omitted if just the label is given. For example,
             ("--num-eggs", "Number of eggs", 2)
        If POSITIONAL, the option prefix (--) is omitted and the option_SPEC
        includes an optional nargs component, such as"
             ("other-files", "Other file names", ["f1", "f2", "f3"], "+")
        """
        # EX: label, _desc, _default = Main.convert_option("--mucho-backflips"); label => "--mucho-backflips"
        ## TEST: result = ["", "", ""]
        opt_label = None
        opt_desc = None
        opt_default = default_value
        opt_nargs = None
        opt_prefix = "--" if not positional else ""
        # TODO: use keyword arguments (or namedtuple)
        if isinstance(option_spec, tuple):
            option_components = list(option_spec)
            opt_label = opt_prefix + option_components[0]
            if len(option_components) > 1:
                opt_desc = option_components[1]
            if len(option_components) > 2:
                opt_default = option_components[2]
            if len(option_components) > 3:
                debug.assertion(positional)
                opt_nargs = option_components[3]
                debug.assertion(positional)
        else:
            opt_label = opt_prefix + tpo.to_string(option_spec)
        debug.assertion(not " " in opt_label)
        result_list = [opt_label, opt_desc, opt_default]
        if positional:
            result_list.append(opt_nargs)
        result = tuple(result_list)
        tpo.debug_format("convert_option({o}, {d}, {p}): self={s} => {r}", 5,
                         o=option_spec, d=default_value, p=positional,
                         s=self, r=result)
        return result

    def convert_argument(self, argument_spec, default_value=None):
        """Convert ARGUMENT_SPEC to (label, description, default) tuple. 
        Note: This is a wrapper around convert_option for positional arguments."""
        debug.trace(6, f"convert_argument({argument_spec}, {default_value}")
        return self.convert_option(argument_spec, default_value, positional=True)

    def get_option_name(self, label):
        """Return internal name for parser options (e.g. dashes converted to underscores)"""
        # EX: dummy_app.get_option_name("mucho-backflips") => "mucho_backflips"
        name = label.replace("-", "_")
        tpo.debug_format("get_option_name({l}) => {n}; self={s}", 6,
                         l=label, n=name, s=self)
        return name

    def has_parsed_option_old(self, label):
        """Whether option for LABEL specified (i.e., non-null value)
        Note: OLD version that checks for non-null value)
        """
        # EX: self.parsed_args = {"it": False}; self.has_parsed_option_old("nonit") => False
        name = self.get_option_name(label)
        has_option = (name in self.parsed_args and (self.parsed_args[name] is not None))
        tpo.debug_format("has_parsed_option_old({l}) => {r}", 6,
                         l=label, r=has_option)
        return has_option

    def has_parsed_option(self, label):
        """Value for LABEL specified or None if not applicable
        Note: This is a deprecated method (use get_parsed_option instead)
        """
        # EX: self.parsed_args = {"it": False}; self.has_parsed_option("notit") => None
        ## TEMP HACK: if called by a subclass, treate as alias to get_parsed_option
        if (self.__class__ != "__main__.Script"):
            debug.trace(4, "Warning: deprecated method: has_parsed_option => get_parsed_option")
            return self.get_parsed_option(label)
        # Return parsed-arg entry for the option
        name = self.get_option_name(label)
        ## TEMP HACK: has_parsed_option returns args.get(opt)
        option_value = self.parsed_args.get(name)
        tpo.debug_format("has_parsed_option({l}) => {r}", 6,
                         l=label, r=option_value)
        return option_value

    def get_parsed_option(self, label, default=None, positional=False):
        """Get value for option LABEL, with dashes converted to underscores. 
        If POSITIONAL specified, DEFAULT value is used if omitted"""
        opt_label = self.get_option_name(label) if not positional else label
        if not self.parsed_args:
            debug.trace(5, "Error: Unexpected condition in get_parsed_option")
            return default
        value = self.parsed_args.get(opt_label)
        # Override null value with default
        if value is None:
            if ((default is None) and ENV_OPTION_PREFIX):
                env_var = f"{ENV_OPTION_PREFIX}_{opt_label}".upper()
                default = system.getenv(env_var)
                if default:
                    debug.trace(4, f"FYI: Using option {label} from env ({opt_label}: {default!r})")
            value = default
            under_label = label.replace("-", "_")
            # Do sanity check for positional argument being checked by mistake
            # TODO: do automatic correction?
            debug.trace_expr(5, label, opt_label, under_label, self.parsed_args)
            if opt_label != label:
                debug.assertion(label not in self.parsed_args)
            elif under_label != label:
                debug.assertion(under_label not in self.parsed_args,
                                f"potential option/argument mismatch for {label}")
            debug.trace_expr(6, label, opt_label, under_label)
        # Return result, after tracing invocation
        tpo.debug_format("get_parsed_option({l}, [{d}], [{p}]) => {v}", 5,
                         l=label, d=default, p=positional, v=value)
        return value

    def get_parsed_argument(self, label, default=None):
        """Get value for positional argument LABEL using DEFAULT value"""
        tpo.debug_format("get_parsed_agument({l}, [{d}])", 6,
                         l=label, d=default)
        ## TODO2: debug.assertion(label in ((l[0] if isinstance(l, list) else l)) for l in self.positional_options)
        is_positional = (label in self.get_arguments(just_positional=True))
        if not is_positional:
            debug.trace(4, f"FYI: Use get_parsed_option for non-positional option {label}")
        return self.get_parsed_option(label, default, positional=is_positional)

    def check_arguments(self, runtime_args):
        """Check command-line arguments
        Note: This uses argparse, which might exit the process
        """
        # Note: Shows env. options when debugging as these are backdoor settings.
        tpo.debug_format("Main.check_arguments({args})", 5, args=runtime_args)
        # TODO: add in detailed usage notes w/ environment option descriptions (see google_word2vec.py)
        if not self.argument_parser:
            self.argument_parser = argparse.ArgumentParser
        usage_notes = self.notes
        ## OLD: if (not usage_notes and SHOW_ENV_OPTIONS):
        if (not usage_notes):
            env_opt_spec = ""
            if (SHOW_ENV_OPTIONS or (f"--{VERBOSE_ARG}" in sys.argv)):
                env_opts = system.formatted_environment_option_descriptions(sort=True, indent=INDENT)
                env_opt_spec = f"- Available env. options:\n{INDENT}{env_opts}"
            elided_path = re.sub(r"^.*/", ".../", sys.argv[0])
            # note: A dash ("-") is used to indicate stdin with filename arg or to bypass usage w/o one
            # TODO1: get dash put in usage to make more explicit, such as in following:
            #     usage: main.py [-h] [--verbose] [filename] [-]
            usage_notes = ("Notes: \n"
                           + ("- Use - for filename to skip usage (i.e., a la stdin).\n" if (not self.skip_input)
                              else "- Use - to skip usage if no arguments needed.\n" if self.auto_help else "")
                           + ("- Use --non-... for any boolean arg\n" if NEGATIVE_BOOL_ARGS else "")
                           + (f"- Use \"ENV1='v1' ENV2='v2' python {elided_path} ...\" for environment options.\n")
                           + env_opt_spec)
        parser = self.argument_parser(description=self.description,
                                      epilog=usage_notes, prog=self.program,
                                      formatter_class=argparse.RawDescriptionHelpFormatter)
        # TODO: use capitalized script description but lowercase argument help

        def add_argument(opt_label, add_short=None, **kwargs):   # pylint: disable=redefined-builtin
            """Wrapper around argparse.ArgumentParser.add_argument
            Note: adds short options string if ADD_SHORT (see self.short_options)"""
            debug.trace(6, f"add_argument{(opt_label, add_short, kwargs)}")
            if add_short is None:
                add_short = self.short_options
            if self.short_options and my_re.search(r"-(-[a-z])[a-z]+", opt_label):
                short_label = my_re.group(1)
                parser.add_argument(short_label, opt_label, **kwargs)
            else:
                parser.add_argument(opt_label, **kwargs)
    
        # Check for options of specific types
        # TODO: consolidate processing for the groups; add option for environment-based default; resolve stupid pylint false positive about unbalanced-tuple-unpacking
        for opt_spec in self.boolean_options:
            (opt_label, opt_desc, opt_default) = self.convert_option(opt_spec, None)    # pylint: disable=unbalanced-tuple-unpacking
            if self.perl_switch_parsing:
                # note: With Perl argument support, booleans treated as integers due to argparse quirk.
                ## TEST: parser.add_argument(opt_label, type=int, nargs="?", default=opt_default, help=opt_desc)
                numeric_default = 1 if opt_default else 0
                add_argument(opt_label, type=int, default=numeric_default, help=opt_desc)
            else:
                add_argument(opt_label, default=opt_default, action="store_true", help=opt_desc)
                if NEGATIVE_BOOL_ARGS:
                    # BAD: label = f"non-{opt_label}"
                    under_label = my_re.sub(r"^__", "", opt_label.replace("-", "_"))
                    label = "--non-" + under_label
                    ## SO-SO
                    ## # note: converts "Description ..." into "Do not description ..."
                    ## opt_desc_uncapitalized = ((opt_desc[:1].lower() + opt_desc[1:]) if opt_desc else "")
                    ## desc = f"Non {opt_desc_uncapitalized}"
                    # note: the argument is not shown in help to avoid clutter
                    desc = argparse.SUPPRESS
                    debug.trace(4, f"Adding negative-boolean: label={label} dest={under_label}")
                    parser.add_argument(label, default=opt_default, dest=under_label, action="store_false", help=desc, add_short=False)
        for opt_spec in self.int_options:
            (opt_label, opt_desc, opt_default) = self.convert_option(opt_spec, None)    # pylint: disable=unbalanced-tuple-unpacking
            add_argument(opt_label, type=int, default=opt_default, help=opt_desc)
        for opt_spec in self.float_options:
            (opt_label, opt_desc, opt_default) = self.convert_option(opt_spec, None)    # pylint: disable=unbalanced-tuple-unpacking
            add_argument(opt_label, type=float, default=opt_default, help=opt_desc)
        for opt_spec in self.text_options:
            (opt_label, opt_desc, opt_default) = self.convert_option(opt_spec, None)    # pylint: disable=unbalanced-tuple-unpacking
            add_argument(opt_label, default=opt_default, help=opt_desc)

        # Add dummy arguments
        # Note: These are used as reminders on how to flesh out the initialization
        if tpo.detailed_debugging():
            if not self.boolean_options:
                parser.add_argument("--TODO-bool-arg", default=False, action="store_true",
                                    help="Add via boolean_options keyword")
            if not self.text_options:
                parser.add_argument("--TODO-text-arg", default="",
                                    help="Add via text_options keyword")
            if not self.int_options:
                parser.add_argument("--TODO-int-arg", type=int, default=0,
                                    help="Add via int_options keyword")
            if not self.float_options:
                parser.add_argument("--TODO-float-arg", default=0.0,
                                    help="Add via float_options keyword")

        # Add positional arguments
        for i, opt_spec in enumerate(self.positional_options):
            (opt_label, opt_desc, opt_default, opt_nargs) = self.convert_argument(opt_spec, "")   # pylint: disable=unbalanced-tuple-unpacking
            # note: a numeric nargs produces a list even if 1, so None used by default
            nargs = opt_nargs
            tpo.debug_format("positional arg {i}, nargs={nargs}", 6, 
                             i=i, nargs=nargs)
            # TODO: add default to opt_desc if not mentioned
            parser.add_argument(opt_label, default=opt_default, nargs=nargs, 
                                help=opt_desc)

        # Add filename last and make optional with "-" default (i.e., for stdin)
        # Note: with nargs=+, the result is a list of filenames (even if one file [WTH?]!)
        if not self.skip_input:
            filename_nargs = ("?" if (not self.multiple_files) else "+")
            tpo.debug_format("filename_nargs={nargs}", 6, nargs=filename_nargs)
            parser.add_argument("filename", nargs=filename_nargs, default="-",
                                help="Input filename")
        elif self.auto_help:
            dash_nargs = "?"
            debug.trace_expr(6, dash_nargs)
            ## TAKE1:
            ## TODO2: hide [dash] from usage (maldito argparse)
            ## parser.add_argument("dash", nargs=dash_nargs, default="-",
            ##                     help="Use - to bypass usage statement")
            parser.add_argument("-", dest="_", help=argparse.SUPPRESS,
                                default="-", action="store_true")
            ## TODO3
            ## parser.add_argument("--", dest="_", help=argparse.SUPPRESS,
            ##                     default="-", action="store_true")
            ## NOTE: currently leads to "error: unrecognized arguments: --" unlike when - used
        else:
            debug.trace(6, "Not adding filename nor dash argument")

        # Optionally, show brief usage and exit
        # note: print_usage just prints command line synopsis (not individual descriptions)
        if (self.brief_usage and (runtime_args == [USAGE_ARG])):
            debug.trace(4, "Just showing (brief) usage and then exiting")
            debug.trace(5, "warning: self.setup won't be invoked")
            parser.print_usage()
            sys.exit()

        # Parse the command line and get result
        tpo.debug_format("parser={p}", 6, p=parser)
        debug.trace_object(8, parser, max_depth=2)
        self.parser = parser
        # note: not trapped to allow for early exit
        self.parsed_args = vars(parser.parse_args(runtime_args))
        debug.trace(5, f"parsed_args = {self.parsed_args}")
        self.verbose = self.get_parsed_option("verbose")

        # Get filename unless input ignored and fixup if returned as list
        # TODO: add an option to retain self.filename as is
        if not self.skip_input:
            self.filename = self.parsed_args["filename"]
            if (isinstance(self.filename, list)):
                if not self.multiple_files:
                    debug.trace(3, "Warning: Making (list) self.filename a string & setting self.other_filenames to remainder")
                file_list = self.filename
                self.other_filenames = file_list[1:]
                self.filename = file_list[0] if len(file_list) else "-"
        debug.trace(6, "end Main.check_arguments()")
        return

    def setup(self):
        """Perform script setup prior to input processing
        Note: This is not invoked if the script exits during --help processing"
        """
        # Note: Use for post-argument proceessing setup
        tpo.debug_format("Main.setup() stub: self={s}", 5, s=self)
        return

    def process_line(self, line):
        """Stub for input processing that just prints the input.
        Note: issues error message about required specialization"""
        # NOTE: the trailing newline is omitted
        # TODO: clarify stripped newline vs. no newline at end of file
        debug.trace_fmt(5, "Main.process_line({l})", l=line)
        if not self.process_line_warning:
            tpo.print_stderr("Warning: specialize process_line")
            self.process_line_warning = True
        print(line)
        return

    def run_main_step(self):
        """Stub for main processing, along with error message"""
        # TODO: use decorator (e.g., @abstract)
        tpo.debug_format("Main.run_main_step(): self={s}", 5, s=self)
        tpo.print_stderr("Internal error: specialize run_main_step")
        return

    def init_input(self):
        """Resolve input stream from either explicit filename or via standard input
        Note: self.newlines is used to override stream (e.g., so \r not treated as line delim)"""
        debug.trace(5, "Main.init_input()")
        self.input_stream = sys.stdin
        if (self.filename and (self.filename != "-")):
            if (isinstance(self.filename, list) or (len(self.other_filenames) > 0)):
                debug.assertion(self.filename != ["-"])
                if not self.multiple_files:
                    # note: check_arguments sets self.other_filenames
                    debug.trace(3, "Warning: Not opening multiple-valued filename arg")
                    debug.trace_expr(3, self.filename, self.other_filenames)
            else:
                debug.assertion(isinstance(self.filename, str))
                debug.assertion(os.path.exists(self.filename))
                mode = ("r" if (not self.binary_input) else "rb")
                self.input_stream = system.open_file(self.filename, mode=mode)
                debug.assertion(self.input_stream)
        if self.newlines:
            debug.trace(4, f"Changing input stream newlines from {self.input_stream.newlines!r} to {self.newlines!r}")
            ## BAD: self.input_stream.newlines = self.newlines
            ## BAD2: sys.stdin.reconfigure(newline=self.newlines)
            self.input_stream = io.TextIOWrapper(self.input_stream.buffer, encoding=self.input_stream.encoding, errors=self.input_stream.errors, newline=self.newlines, line_buffering=self.input_stream.line_buffering, write_through=self.input_stream.write_through)
            debug.trace_object(4, self.input_stream)
    
    def run(self):
        """Runner for script processing"""
        tpo.debug_print("Main.run()", 5)
        # TODO: decompose (e.g., isolate input proecessing)

        # Have client do pre-input initialization (e.g., argument extraction)
        self.setup()

        # Initiate the main input processing
        self.init_input()
        try:
            # If not automatic input, process the main step of script
            if self.manual_input:
                self.run_main_step()
            # Otherwise have client process input line by line
            else:
                # TODO: Trace status only if script blocks waiting for user
                debug.trace(2, "Processing input")
                self.process_input()
        except BrokenPipeError:
            ## TODO: exit gracefully (e.g., after wrap_up)
            ## debug.trace_fmt(6, "Silly exception processing input: {exc}", 
            ##                 exc=system.get_exception)
            ##
            ## exit("Note: you can Ignore any silly BrokenPipeError"s thrown by Python!")
            ##
            ## pass
            ##
            ## take 3+:
            ## Based on https://stackoverflow.com/questions/26692284/how-to-prevent-brokenpipeerror-when-doing-a-flush-in-python
            ## sys.stderr.close()                       [doesn"t work for Python 3]
            ##
            ## take 4 [gotta hate Python!]:
            # Python flushes standard streams on exit; redirect remaining output
            # to devnull to avoid another BrokenPipeError at shutdown
            devnull = os.open(os.devnull, os.O_WRONLY)
            os.dup2(devnull, sys.stdout.fileno())
            sys.exit(1)  # Python exits with error code 1 on EPIPE

        # Invoke client end processing
        self.wrap_up()

        # Remove any temporary files
        self.clean_up()
        return

    def read_entire_input(self):
        """Returns all input (either from specified filename or stdin)
        Notes:
        - This is simple alternative to the read_input generator intended for use with dummy_app
        - Another alternative is to invoke self.init_input and then use self.input_stream
        """
        debug.trace(5, "Main.read_entire_input()")
        self.init_input()
        debug.trace(2, "Processing entire input")
        debug.trace_object(4, self.input_stream)
        input_text = self.input_stream.read()
        debug.trace_expr(6, input_text)
        return input_text
    
    def read_input(self):
        """Generator for producing lines of text from the input (without newlines).
        Notes:
        0. Use read_entire_input for method to return all text at once (i.e., non-generator).
        1. This is called automatically via process_input.
        2. When page mode is in effect, 
           a. lines don"t include form feeds
           b. pages are not buffered: the page number is just tracked.
        3. Paragraph end at empty lines or form feeds (i.e., implicit).
        """
        # TODO: add special page-input mode
        # Note: para_num reset here and updated by process_input
        tpo.debug_format("Main.read_input(): {input}", 5,
                         input=self.input_stream)

        # Optionally return input all at once
        # Note: Final newline is removed, as this feeds into process_line,
        # which there was one
        if self.file_input_mode:
            contents = self.input_stream.read()
            if contents.endswith("\n"):
                contents = contents[:-1]
            debug.trace_fmt(6, "yielding entire file [Par1/L1]: {c}",
                            c=contents)
            yield contents
            return

        # Start line-based input
        self.page_num = 1
        self.para_num = 1
        self.rel_para_num = 1
        self.line_num = 0
        self.rel_line_num = 0
        self.char_offset = 0
        for line in self.input_stream:
            self.rel_line_num += 1
            self.line_num += 1
            self.raw_line = line
            if line.endswith("\n"):
                line = line[:-1]
            debug.trace_fmt(6, "L{n}: {l}", n=self.line_num, l=line)
            if self.force_unicode:
                line = tpo.ensure_unicode(line)
            ## TEST: debug.trace(7, f"\ttype(line): {type(line)}; offset={self.input_stream.tell()}")
            if self.track_pages:
                for i, line_segment in enumerate(line.split(FORM_FEED)):
                    debug.trace(7, f"LS{i}: {line_segment}")
                    self.end_of_page = False
                    if i == 0:
                        self.end_of_page = (line != line_segment)
                    else:
                        self.end_of_page = True
                        if RETAIN_FORM_FEED:
                            line_segment = FORM_FEED + line_segment
                    ## OLD: if line_segment and (i > 0):
                    ## NEW:
                    if (i > 0):
                        self.line_num += 1
                        self.rel_line_num += 1
                        if self.end_of_page:
                            self.page_num += 1
                            self.rel_para_num = 1
                            self.rel_line_num = 1
                    ## OLD: if line_segment:
                    ## NEW:
                    if True:            # pylint: disable=using-constant-test
                        debug.trace_fmt(6, "yielding line segment [Pg{pg}/Par{par}/L{ln}]: {ls}",
                                        pg=self.page_num, par=self.rel_para_num, ln=self.rel_line_num, ls=line_segment)
                        yield line_segment
                    self.char_offset += len(line_segment)
                    debug.trace_expr(7, self.page_num)
                if (line != self.raw_line):
                    self.char_offset += 1
            else:
                debug.trace_fmt(6, "yielding line [Par{par}/L{lnum}]: {l}",
                                par=self.rel_para_num, lnum=self.rel_line_num, l=line)
                yield line
                self.char_offset += len(self.raw_line)
        return

    def is_line_mode(self):
        """Whether processing normal lines (not paragraphs or entire files)"""
        return  (not (self.paragraph_mode or self.file_input_mode))

    def process_input(self):
        """Process each line in current input stream (or stdin):
        Note: if paragraph mode enabled the input is processed in groups of lines separated by an entirely blank line (i.e., length is 0)"""
        # Note: self.raw_line can be used to check for missing newline at end of file
        tpo.debug_format("Main.process_input(): {input}", 5,
                         input=self.input_stream)
        self.rel_line_num = 0
        if self.paragraph_mode:
            self.para_num = 0
        paragraph = ""
        last_line = None
        line_mode = self.is_line_mode()
        debug.assertion(debug.xor3(line_mode, self.paragraph_mode, self.file_input_mode))

        # Read next line (or line segment if in page mode and form feed in line)
        for line in self.read_input():
            # Process as is if in regular line mode
            if (line_mode or self.file_input_mode):
                self.process_line(line)
                if line_mode:
                    debug.assertion("\n" not in line)

            # Process non-empty lines grouped together if in paragraph mode.
            # Notes:
            # - Paragraphs can include multiple, trailing newlines, so that the
            #   client programs see the entire input text.
            # - Final newline is removed, as fed into process_line.
            # TODO: Have option to model perl-style paragraph mode more closely
            # with respect to handling more than 2 newlines between paragraphs; see
            #     https://perldoc.perl.org/variables/$/.
            else:
                new_paragraph = None
                if self.end_of_page:
                    new_paragraph = (line + "\n")
                    paragraph = ""
                elif ((last_line == "") and line):
                    new_paragraph = paragraph
                    paragraph = (line + "\n")
                else:
                    paragraph += (line + "\n")
                debug.trace_expr(7, new_paragraph, paragraph)
                if new_paragraph:
                    self.rel_para_num += 1
                    self.para_num += 1
                    debug.assertion(new_paragraph.endswith("\n\n") or (new_paragraph == "\n"))
                    if new_paragraph.endswith("\n"):
                        new_paragraph = new_paragraph[:-1]
                    self.process_line(new_paragraph)
            debug.assertion(not (self.track_pages and (not RETAIN_FORM_FEED) and (FORM_FEED in line)))
            last_line = line

        # Process the last set of lines if in paragraph mode
        # Note: Final newline is removed (as per process_line).
        if (self.paragraph_mode and paragraph):
            self.rel_para_num += 1
            self.para_num += 1
            debug.trace(5, "processing last paragraph")
            debug.assertion(paragraph.endswith("\n") or (paragraph == ""))
            if paragraph.endswith("\n"):
                paragraph = paragraph[:-1]
            self.process_line(paragraph)

        return

    def wrap_up(self):
        """Default end processing"""
        tpo.debug_format("Main.wrap_up() stub: self={s}", 5, s=self)
        return

    def clean_up(self):
        """Removes temporary files, etc."""
        # note: not intended to be overridden
        tpo.debug_format("Main.clean_up(): self={s}", 5, s=self)
        if not KEEP_TEMP_FILES:
            ## TODO2: 3=>6
            debug.trace_fmt(4, "Deleting any temporary files: {files}",
                            files=gh.run(f"echo {self.temp_base}* {self.temp_file}* | sort -u"))
                                
            # Remove all temp_base* files (or the temp_base directory)
            if self.use_temp_base_dir:
                gh.run("rm -rvf {dir}", dir=self.temp_base)
            else:
                gh.run("rm -vf {file}*", file=self.temp_base)
            # Likewise remove all temp_file* files
            if (self.temp_file != self.temp_base):
                gh.run("rm -vf {file}*", file=self.temp_file)
        return

#-------------------------------------------------------------------------------
# Global instance for convenient adhoc usage
# 
# note: useful for temporary file support (e.g., dummy_app.temp_file)
#

dummy_app = Main([])

debug.trace_current_context(8, "main.py context")

#------------------------------------------------------------------------

if __name__ == "__main__":
    system.print_stderr(f"Warning: {__file__} is not intended to be run standalone\n")
    # note: Follwing used for argument parsing
    main = Main(description=__doc__)
    main.run()




#! /usr/bin/env python
#
# Convenience class for regex searching, providing simple wrapper around
# static match results.
#
# Note: This provides tracing for commonly used functions (e.g., search and sub),
# with aliasing used for miscellaneous others (e.g., subn).
#
# Example usage:
#
#    from my_regex import my_re
#    ...
#    if (my_re.search(r"^(\d+)\:?(\d*)\s+(\d+)\:?(\d*)\s+\S.*", line)):
#        (start_hours, start_mins, end_hours, end_mins) = my_re.groups()
#
#--------------------------------------------------------------------------------
# TODO:
# - Add examples for group(), groups(), etc.
# - Clean up script (e.g., regex => regex_wrapper).
# - Add perl-inspired accessors (e.g., PREMATCH, POSTMATCH).
#

"""Wrapper class for regex match results"""

# Standard packages
import re
## TODO: from re import *

# Installed packages
import six

# Local packages
from mezcla import debug
from mezcla import system

# Expose public symbols from re package, plus the wrapper class and global instance
## OLD:
## # Expose public symbols from re package,
## __all__ = re.__all__
## DEBUG: system.print_error("checking SKIP_RE_ALL")
RE_ALL = (not system.getenv_bool("SKIP_RE_ALL", False,
                                 "Don't use re.__all__: for sake of pylint"))
__all__ = ['regex_wrapper', 'my_re']
if RE_ALL:
    ## TODO: __all__ = re.__all__ + ['regex_wrapper', 'my_re']
    __all__ += re.__all__
    pass
else:
    debug.trace(4, "Omitting use of re __all__")

REGEX_TRACE_LEVEL = system.getenv_int("REGEX_TRACE_LEVEL", debug.QUITE_DETAILED,
                                      "Trace level for my_regex")
    
## TODO # HACK: make sure regex can be used as plug-in replacement 
## from from re import *

## TEST: Attempts to work around Python enum extension limitation
##
## OLD: class regex_wrapper(object):
## HACK: inherit from RegexFlag, so pylint not confused by attrib copying
## NOTE: Maldito python!
##    class regex_wrapper(re.RegexFlag):
##    => TypeError: Cannot extend enumerations
## via https://stackoverflow.com/questions/33679930/how-to-extend-python-enum:
##    Subclassing an enumeration is allowed only if the enumeration does not define any members.
##
##
## HACK2: try via multiple inheritance
##
## class TraceLevel(object):
##     """Simple class with trace level
##     Note: this is just used to work around issue subclassing enums"""
##     # TODO: rework trace level in debug module to be class based
##     TRACE_LEVEL = debug.QUITE_DETAILED
##
## class regex_wrapper(TraceLevel, re.RegexFlag):
##     ...
##
## 
## class MalditoPython(re.RegexFlag):
##     """Just what it says"""
##     pass
##

class regex_wrapper():
    """Wrapper class over re to implement regex search that saves match results
    note: Allows regex to be used directly in conditions"""
    # TODO: IGNORECASE = re.IGNORECASE, etc.
    # import from RE so other methods supported directly (and above constants)
    TRACE_LEVEL = REGEX_TRACE_LEVEL
    ##
    ## Malditos python & pylint!
    ASCII = re.ASCII
    IGNORECASE = re.IGNORECASE
    LOCALE = re.LOCALE
    MULTILINE = re.MULTILINE
    DOTALL = re.DOTALL
    VERBOSE = re.VERBOSE
    UNICODE = re.UNICODE
    # TODO: add miscellaneous re functions (e.g., subn)
    
    # pylint: disable=super-init-not-called
    #
    def __init__(self, ):
        debug.trace_fmtd(4, "my_regex.__init__(): self={s}", s=self)
        self.match_result = None
        # TODO: self.regex = ""

        # HACK: Import attributes from re class
        # TODO3: see if clean way to do this
        # TODO4: find way to disable pylint no-member warning
        try:
            for var in re.__all__:
                if var not in dir(self):
                    debug.trace(9, f"Copying {var} from re into {self}")
                    setattr(self, var, getattr(re, var))
        except:
            system.print_exception_info("__init__ re.* importation")

    def check_pattern(self, regex):
        """Apply sanity checks to REGEX when debugging
        Note: Added to account for potential f-string confusion"""
        # TODO: Add way to disable check
        debug.reference_var(self)
        check_regex = r"[^{]\{[^0-9][A-Fa-f0-9]*[^{}]+\}[^}]"
        if isinstance(regex, bytes):
            check_regex = check_regex.encode()
        if debug.debugging(1):
            match = re.search(check_regex, regex)
            if match:
                system.print_error(f"Warning: potentially unresolved f-string in {regex} at {match.start(0)}")

    def search(self, regex, text, flags=0, base_trace_level=None):
        """Search for REGEX in TEXT with optional FLAGS and BASE_TRACE_LEVEL (e.g., 6)"""
        ## TODO: rename as match_anywhere for clarity
        if base_trace_level is None:
            base_trace_level = self.TRACE_LEVEL
        debug.trace_fmtd((1 + base_trace_level), "my_regex.search({r!r}, {t!r}, {f}): self={s}",
                         r=regex, t=text, f=flags, s=self)
        ## OLD: debug.assertion(isinstance(text, six.string_types))
        debug.assertion(isinstance(text, (str, bytes)) and (type(regex) == type(text)))
        self.check_pattern(regex)
        self.match_result = re.search(regex, text, flags)
        if self.match_result:
            debug.trace_fmt(base_trace_level, "match: {m!r}; regex: {r!r}", m=self.grouping(), r=regex)
        return self.match_result

    def match(self, regex, text, flags=0, base_trace_level=None):
        """Match REGEX to TEXT with optional FLAGS and BASE_TRACE_LEVEL (e.g., 6)"""
        ## TODO: rename as match_start for clarity; add match_all method (wrapper around fullmatch)
        if base_trace_level is None:
            base_trace_level = self.TRACE_LEVEL
        debug.trace_fmtd((1 + base_trace_level), "my_regex.match({r!r}, {t!r}, {f}): self={s}",
                         r=regex, t=text, f=flags, s=self)
        self.check_pattern(regex)
        self.match_result = re.match(regex, text, flags)
        if self.match_result:
            debug.trace_fmt(base_trace_level, "match: {m!r}; regex: {r!r}", m=self.grouping(), r=regex)
        return self.match_result

    def get_match(self):
        """Return match result object for last search or match"""
        result = self.match_result
        debug.trace_fmtd(self.TRACE_LEVEL, "my_regex.get_match() => {r!r}: self={s}",
                         r=result, s=self)
        return result

    def group(self, num):
        """Return group NUM from match result from last search"""
        debug.assertion(self.match_result)
        result = self.match_result and self.match_result.group(num)
        debug.trace_fmtd(self.TRACE_LEVEL, "my_regex.group({n}) => {r!r}: self={s}",
                         n=num, r=result, s=self)
        return result

    def groups(self):
        """Return all groups in match result from last search"""
        debug.assertion(self.match_result)
        result = self.match_result and self.match_result.groups()
        debug.trace_fmt(self.TRACE_LEVEL, "my_regex.groups() => {r!r}: self={s}",
                        r=result, s=self)
        return result

    def grouping(self):
        """Return groups for match result or entire matching string if no groups defined"""
        # Note: this is intended to facilitate debug tracing; see example in search method above
        result = self.match_result and (self.match_result.groups() or self.match_result.group(0))
        debug.trace_fmt(self.TRACE_LEVEL + 1, "my_regex.grouping() => {r!r}: self={s}", r=result, s=self)
        return result

    def start(self, group=0):
        """Start index for GROUP"""
        result = self.match_result and self.match_result.start(group)
        debug.trace_fmt(self.TRACE_LEVEL + 1, "my_regex.start({g}) => {r!r}: self={s}", r=result, s=self, g=group)
        return result

    def end(self, group=0):
        """End index for GROUP"""
        result = self.match_result and self.match_result.end(group)
        debug.trace_fmt(self.TRACE_LEVEL + 1, "my_regex.end({g}) => {r!r}: self={s}", r=result, s=self, g=group)
        return result

    def sub(self, pattern, replacement, string, *, count=0, flags=0):
        """Version of re.sub requiring explicit keyword parameters"""
        # Note: Explicit keywords enforced to avoid confusion
        result = re.sub(pattern, replacement, string, count, flags)
        debug.reference_var(self)
        debug.trace(self.TRACE_LEVEL + 1, f"my_regex.sub({pattern!r}, {replacement!r}, {string!r}, [count=[count]], flags={flags}]) => {result!r}\n")
        self.check_pattern(pattern)
        return result

    def span(self, group=0):
        """Tuple with GROUP start and end"""
        return (self.match_result and self.match_result.span(group))

    def split(self, pattern, string, maxsplit=0, flags=0):
        """Use PATTERN to split STRING, optionally up to MAXSPLIT with FLAGS"""
        result = re.split(pattern, string, maxsplit, flags)
        debug.trace_fmt(self.TRACE_LEVEL, "split{args} => {r!r}",
                        args=tuple([pattern, string, maxsplit, flags]), r=result)
        return result
    
    def findall(self, pattern, string, flags=0):
        """Use PATTERN to split STRING, optionally with specified FLAGS"""
        result = re.findall(pattern, string, flags)
        debug.trace_fmt(self.TRACE_LEVEL, "findall{args} => {r!r}",
                        args=tuple([pattern, string, flags]), r=result)
        return result

    def escape(self, text):
        """Escape special characters in TEXT"""
        result = re.escape(text)
        debug.trace(self.TRACE_LEVEL + 1, f"escape({text!r}) => {result!r}")
        return result
    
#...............................................................................
# Initialization
#
# note: creates global instance for convenience (and backward compatibility)

my_re = regex_wrapper()

if __name__ == '__main__':
    system.print_error("Warning: not intended for command-line use")
    ## Note: truth in advertising:
    ## debug.trace(4, f"mp: {MalditoPython()}")



#! /usr/bin/env python
#
# Test(s) for ../file_utils.py
#
# Notes:
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_file_utils.py
#

"""Tests for file_utils module"""

# Standard packages
import json
import re

# Installed packages
import pytest

# Local packages
from mezcla.unittest_wrapper import TestWrapper
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla import system

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	                global module object
#    TestTemplate.script_module:        path to file
import mezcla.file_utils as THE_MODULE

class TestFileUtils(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.derive_tested_module_name(__file__)
    use_temp_base_dir = True

    def test_get_directory_listing(self):
        """
        Tests for get_directory_listing(path          = 'PATH',
                                        recursive     = False,
                                        long          = False,
                                        readable_size = False,
                                        return_string = True,
                                        make_unicode  = False)
        """

        # Setup files and folders
        folders     = [self.temp_base, gh.form_path(self.temp_base, 'other_folder')]
        filenames   = ['analize.py', 'debug.cpp', 'main.txt', 'misc_utils.perl']

        for foldername in folders:
            ## TODO3: use mezcla wrappers (e.g., gh.full_mkdir)
            gh.run(f'mkdir {foldername}')
            for filename in filenames:
                gh.run(f'touch {gh.form_path(foldername, filename)}')

        # Run Test
        list_result = THE_MODULE.get_directory_listing(f'{self.temp_base}', recursive=True, long=True, return_string=True)

        for line in list_result:
            assert bool(re.search(r"[drwx-]+\s+\d+\s+\w+\s+\w+\s+\d+\s+\w+\s+\d+\s+\d\d:\d\d\s+[\w/]+", line))

    def test_get_information(self):
        """Tests for def get_information(path,
                                         readable = False,
                                         return_string = False)
        """
        gh.run(f'touch {self.temp_file}')

        ls_result = gh.run(f'ls -l {self.temp_file}')
        ls_result = re.sub(r'\s+', ' ', ls_result)

        assert THE_MODULE.get_information(self.temp_file, return_string=True).lower() == ls_result.lower()

    def test_get_permissions(self):
        """Tests for get_permissions(path)"""

        # Setup
        test_file = gh.form_path(self.temp_base, 'some-file.cpp')
        gh.run(f'touch {test_file}')

        # Run
        assert THE_MODULE.get_permissions(test_file) == gh.run(f'ls -l {test_file}')[:10]
        assert THE_MODULE.get_permissions(self.temp_base) == gh.run(f'ls -ld {self.temp_base}')[:10]

    def test_get_modification_date(self):
        """Tests for get_modification_date(path)"""
        gh.run(f'touch {self.temp_file}')
        ls_date = gh.run(f'ls -l {self.temp_file}').lower()
        # Extract date
        ls_date = re.search(r'\w\w\w +\d\d? +\d\d:\d\d', ls_date).group()
        # Remove extra spaces
        ls_date = re.sub(r'\s+', ' ', ls_date)
        assert THE_MODULE.get_modification_date(self.temp_file, strftime='%b %-d %H:%M').lower() == ls_date

    def test_json_to_jsonl(self):
        """Tests for json_to_jsonl"""
        in_file = gh.form_path(self.temp_base, "some-file.json")
        out_file = gh.form_path(self.temp_base, "some-file.jsonl")
        sample_array = [1, 2, 3]
        system.write_file(in_file, f"{sample_array}\n")
        THE_MODULE.json_to_jsonl(in_file, out_file)
        assert(system.read_lines(out_file) == list(map(str, sample_array)))

    def test_jsonl_to_json(self):
        """Tests for jsonl_to_json"""
        in_file = gh.form_path(self.temp_base, "another-file.jsonl")
        out_file = gh.form_path(self.temp_base, "another-file.json")
        sample_array = ["dog", "cat", {"fav": "???"}]
        system.write_lines(in_file, [json.dumps(item) for item in sample_array])
        THE_MODULE.jsonl_to_json(in_file, out_file)
        assert(json.loads(system.read_file(out_file)) == sample_array)


#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python
#
# Test(s) for ../keras_param_search.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_keras_param_search.py
#

"""Tests for keras_param_search module"""

# Standard packages
## TODO: import re

# Installed packages
import pytest

# Local packages
from mezcla import debug

# Load module conditionally because not part of default installation.
# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
try:
    import mezcla.keras_param_search as THE_MODULE
except:
    THE_MODULE = None
    debug.trace_exception(1, "keras_param_search import")

class TestKerasParamSearch:
    """Class for testcase definition"""

    @pytest.mark.skipif(not THE_MODULE, reason="Unable to load module")
    def test_round3(self):
        """Ensure round3 works as expected"""
        debug.trace(4, f"test_round3(); self={self}")
        assert THE_MODULE.round3(1.0001) == 1.000
        assert THE_MODULE.round3(1.001) == 1.001

    @pytest.mark.xfail
    def test_non_negative(self):
        """Ensure non_negative works as expected"""
        debug.trace(4, f"test_non_negative(); self={self}")
        ## TODO: WORK-IN=PROGRESS
        assert(False)

    @pytest.mark.xfail
    def test_create_feature_mapping(self):
        """Ensure create_feature_mapping works as expected"""
        debug.trace(4, f"test_create_feature_mapping(); self={self}")
        assert THE_MODULE.create_feature_mapping(['c', 'b', 'b', 'a']) == {'c':0, 'b':1, 'a':2}

    @pytest.mark.xfail
    def test_create_keras_model(self):
        """Ensure create_keras_model works as expected"""
        debug.trace(4, f"test_create_keras_model(); self={self}")
        ## TODO: WORK-IN=PROGRESS
        assert(False)

    ## TODO: test MyKerasClassifier class
    ## TODO: test main


if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python
#
# Example of how to plot multiclass precision/recall curves with sklearn
# (via "binarizing" of output).
#
# Notes:
# - Based on https://stackoverflow.com/questions/56090541/how-to-plot-precision-and-recall-of-multiclass-classifier.
# - Also see https://scikit-learn.org/1.0/auto_examples/model_selection/plot_precision_recall.html.
#
# TODO:
# - Have option to take confusion matrix as input (e.g., to reproduce Vertex AI results).
#...............................................................................
# Original example from Stack Overflow:
#
# 1. general settings, learning and prediction
# 
#   from sklearn.datasets import fetch_mldata
#   from sklearn.model_selection import train_test_split
#   from sklearn.ensemble import RandomForestClassifier
#   from sklearn.multiclass import OneVsRestClassifier
#   from sklearn.metrics import precision_recall_curve, roc_curve
#   from sklearn.preprocessing import label_binarize
#
#   import matplotlib.pyplot as plt
#   #%matplotlib inline
#
#   mnist = fetch_mldata("MNIST original")
#   n_classes = len(set(mnist.target))
#
#   Y = label_binarize(mnist.target, classes=[*range(n_classes)])
#
#   X_train, X_test, y_train, y_test = train_test_split(mnist.data,
#                                                       Y,
#                                                       random_state = 42)
#
#   clf = OneVsRestClassifier(RandomForestClassifier(n_estimators=50,
#                                max_depth=3,
#                                random_state=0))
#   clf.fit(X_train, y_train)
#
#   y_score = clf.predict_proba(X_test)
# 2. precision-recall curve
# 
#   # precision recall curve
#   precision = dict()
#   recall = dict()
#   for i in range(n_classes):
#       precision[i], recall[i], _ = precision_recall_curve(y_test[:, i],
#                                                           y_score[:, i])
#       plt.plot(recall[i], precision[i], lw=2, label='class {}'.format(i))
#
#   plt.xlabel("recall")
#   plt.ylabel("precision")
#   plt.legend(loc="best")
#   plt.title("precision vs. recall curve")
#   plt.show()
#   enter image description here
#
# 3. ROC curve
#
#   # roc curve
#   fpr = dict()
#   tpr = dict()
#
#   for i in range(n_classes):
#       fpr[i], tpr[i], _ = roc_curve(y_test[:, i],
#                                     y_score[:, i]))
#       plt.plot(fpr[i], tpr[i], lw=2, label='class {}'.format(i))
#
#   plt.xlabel("false positive rate")
#   plt.ylabel("true positive rate")
#   plt.legend(loc="best")
#   plt.title("ROC curve")
#   plt.show()
#
#...............................................................................
#
"""Example from https://stackoverflow.com/questions/56090541/how-to-plot-precision-and-recall-of-multiclass-classifier"""

# Standard packages
import re

# Installed packages
import pandas as pd
from sklearn.datasets import fetch_openml
from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestClassifier
from sklearn.multiclass import OneVsRestClassifier
from sklearn.metrics import precision_recall_curve, roc_curve
from sklearn.preprocessing import label_binarize
#
from mezcla import data_utils as du
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
from mezcla import system
from mezcla import text_utils

# Constants
DATASET_NAME = system.getenv_text("DATASET_NAME", "mnist_784",
                                  ## TODO: "MNIST ???" (see https://www.openml.org/d/554)
                                  "Name of OpenML dataset or basename of local CSV file")
USE_OPENML_DEFAULT = not system.file_exists(DATASET_NAME)
USE_OPENML = system.getenv_bool("USE_OPENML", USE_OPENML_DEFAULT,
                                "Whether DATASET_NAME refers to OpenML dataset")
SKIP_PLOT = system.getenv_bool("SKIP_PLOT", False,
                               "Don't show plot: just output data")
SHOW_PLOT = not SKIP_PLOT
INTERACTIVE_PLOT = system.getenv_bool("INTERACTIVE_PLOT", False,
                                      "Use interactive plot via matplotlib.plot")
DEFAULT_BASENAME = re.sub(r".py\w*$", "", gh.basename(__file__))
BASENAME = system.getenv_text("BASENAME", DEFAULT_BASENAME,
                              "Basename for about files")
CLASS_VAR = system.getenv_value("CLASS_VAR", None,
                                "Name of classification variable if not last")
CLASS_LABELS = system.getenv_value("CLASS_LABELS", None,
                                   "Comma-separated list of class labels")

# Optional packages
if SHOW_PLOT:
    import matplotlib.pyplot as plt
    ## TODO: #%matplotlib inline

## TODO: Constants for switches omitting leading dashes (e.g., DEBUG_MODE = "debug-mode")
## Note: Run following in Emacs to interactively replace TODO_ARG with option label
##    M-: (query-replace-regexp "todo\\([-_]\\)arg" "arg\\1name")
## where M-: is the emacs keystroke short-cut for eval-expression.
TODO_ARG = "TODO-arg"
## ALT_TODO_ARG = "alt-todo-arg"
## TODO_FILENAME = "TODO-filename"

## TODO:
## # Environment options
## # Note: These are just intended for internal options, not for end users.
## # It also allows for enabling options in one place rather than four
## # (e.g., [Main member] initialization, run-time value, and argument spec., along
## # with string constant definition).
## #
## FUBAR = system.getenv_bool("FUBAR", False,
##                            description="Fouled Up Beyond All Recognition processing")

#-------------------------------------------------------------------------------
# Utility functions

def reset_plot():
    """Reset the mathplotlib state"""
    # Note: See https://stackoverflow.com/questions/741877/how-do-i-tell-matplotlib-that-i-am-done-with-a-plot
    # TODO: Make sure no other state retained

    # Clear the axes
    plt.cla()

    # Clear the figure
    plt.clf()
    return

#-------------------------------------------------------------------------------

class Script(Main):
    """Input processing class"""
    # TODO: -or-: """Adhoc script class (e.g., no I/O loop, just run calls)"""
    ## TODO: class-level member variables for arguments (avoids need for class constructor)
    todo_arg = False
    ## alt_todo_arg = ""

    # TODO: add class constructor if needed for non-standard initialization
    ## def __init__(self, *args, **kwargs):
    ##     debug.trace_fmtd(5, "Script.__init__({a}): keywords={kw}; self={s}",
    ##                      a=",".join(args), kw=kwargs, s=self)
    ##     super(Script, self).__init__(*args, **kwargs)

    def setup(self):
        """Check results of command line processing"""
        debug.trace_fmtd(5, "Script.setup(): self={s}", s=self)
        ## TODO: extract argument values
        self.todo_arg = self.get_parsed_option(TODO_ARG, self.todo_arg)
        ## self.alt_todo_arg = self.get_parsed_option(alt_todo_arg, self.alt_todo_arg)
        # TODO: self.TODO_filename = self.get_parsed_argument(TODO_FILENAME)
        debug.trace_object(5, self, label="Script instance")
        return

    def run_main_step(self):
        """Main processing step"""
        debug.trace_fmtd(5, "Script.run_main_step(): self={s}", s=self)

        # 1. general settings, learning and prediction
        feature_data = target_data = None
        if USE_OPENML:
            # pylint: disable=no-member
            dataset = fetch_openml(system.quote_url_text(DATASET_NAME))
            feature_data = dataset.data
            target_data = dataset.target
        else:
            dataset_filename = DATASET_NAME if system.file_exists(DATASET_NAME) else (DATASET_NAME + ".csv")
            dataset = du.read_csv(dataset_filename)
            # TODO: add CLASS_POS?
            class_var = CLASS_VAR if CLASS_VAR else dataset.columns[-1]
            # pylint: disable=no-member
            feature_data = dataset.copy()
            feature_data.drop(class_var, axis=1)
            ## OLD: class_index = list(dataset.columns).index(class_var)
            ## BAD: features_indices = dataset.columns[0:class_index] + dataset.columns[class_index + 1:]
            ## BAD2: features_indices = dataset.columns[0:class_index]
            ## BAD2: features_indices.append(dataset.columns[class_index + 1:])
            ## BAD3: features_indices = pd.concat(dataset.columns[0:class_index], dataset.columns[class_index + 1:])
            ## OLD: debug.assertion(class_index == 0)
            ## OLD: features_indices = dataset.columns[class_index + 1:]
            ## OLD2: target_indices = dataset.columns[class_index: class_index + 1]
            ## BAD: target_data = dataset[target_indices].iloc[0]
            target_data = dataset[class_var]
        debug.assertion(isinstance(feature_data, pd.DataFrame))
        debug.assertion(isinstance(target_data, pd.Series))
        n_classes = len(set(target_data))
        ## BAD: Y = label_binarize(target_data, classes=[*range(n_classes)])
        class_labels = target_data.unique()
        Y = label_binarize(target_data, classes=class_labels)
        debug.assertion(not all(list((sum(y) == 0) for y in Y)))
        X_train, X_test, y_train, y_test = train_test_split(feature_data, Y,
                                                            random_state=42)
        debug.trace_expr(4, dataset, n_classes, Y)
        debug.trace_expr(5, X_train, X_test, y_train, y_test)
        # TODO: parameterize (e.g., classifier name and common options)
        clf = OneVsRestClassifier(RandomForestClassifier(n_estimators=50, 
                                                         max_depth=3,
                                                         random_state=0))
        clf.fit(X_train, y_train)
        y_score = clf.predict_proba(X_test)
        debug.trace_object(5, clf)
        debug.trace_expr(6, y_score)

        # 2. precision-recall curve
        reset_plot()
        precision = dict()
        recall = dict()
        thresholds = dict()
        plot_class_labels = class_labels
        if CLASS_LABELS:
            plot_class_labels = text_utils.extract_string_list(CLASS_LABELS)
        for i in range(n_classes):
            precision[i], recall[i], thresholds[i] = precision_recall_curve(y_test[:, i],
                                                                            y_score[:, i])
            plt.plot(recall[i], precision[i], linewidth=2, label=plot_class_labels[i])
            debug.trace_expr(6, precision[i], recall[i], thresholds[i])
        #
        plt.xlabel("recall")
        plt.ylabel("precision")
        plt.legend(loc="best")
        plt.title("precision vs. recall curve")
        debug.trace_object(5, plt)
        output_files = []
        if INTERACTIVE_PLOT:
            plt.show()
        else:
            output_files.append(BASENAME + ".pr-plot.png")
            plt.savefig(output_files[-1])

        # 3. ROC curve
        reset_plot()
        fpr = dict()
        tpr = dict()
        for i in range(n_classes):
            fpr[i], tpr[i], thresholds[i] = roc_curve(y_test[:, i],
                                                      y_score[:, i])
            plt.plot(fpr[i], tpr[i], lw=2, label=plot_class_labels[i])
            debug.trace_expr(6, precision[i], recall[i], thresholds[i])
        #
        plt.xlabel("false positive rate")
        plt.ylabel("true positive rate")
        plt.legend(loc="best")
        plt.title("ROC curve")
        debug.trace_object(5, plt)
        if INTERACTIVE_PLOT:
            plt.show()
        else:
            output_files.append(BASENAME + ".roc-plot.png")
            plt.savefig(output_files[-1])

        # Output status
        if not INTERACTIVE_PLOT:
            print("Output files:\n\t{files}\n\t".format(files="\n\t".join(output_files)))

        return


#-------------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context(level=debug.QUITE_DETAILED)
    debug.trace_fmt(4, "Environment options: {eo}",
                    eo=system.formatted_environment_option_descriptions())
    app = Script(
        description=__doc__,
        # Note: skip_input controls the line-by-line processing, which is inefficient but simple to
        # understand; in contrast, manual_input controls iterator-based input (the opposite of both).
        skip_input=False,
        manual_input=True,
        # TODO: Disable inference of --help argument
        ## auto_help=False,
        ## TODO: specify options and (required) arguments
        ## TODO: boolean_options=[(TODO_ARG, "TODO-desc")],
        # Note: FILENAME is default argument unless skip_input
        ## positional_arguments=[FILENAME1, FILENAME2], 
        ## text_options=[(alt_todo_arg, "TODO-desc")],
        # Note: Following added for indentation: float options are not common
        float_options=None)
    app.run()




#! /usr/bin/env python
#
# Test(s) for ../rgb_color_name.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_rgb_color_name.py
#

"""Tests for rgb_color_name module"""

# Standard packages
import re

# Installed packages
import pytest

# Local packages
from mezcla.unittest_wrapper import TestWrapper
from mezcla import debug
from mezcla import system
from mezcla import glue_helpers as gh

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.rgb_color_name as THE_MODULE
#
# Note: sanity test for customization (TODO: remove if desired)
if not re.search(__file__, r"\btemplate.py$"):
    debug.assertion("mezcla.template" not in str(THE_MODULE))

## TODO:
## # Environment options
## # Note: These are just intended for internal options, not for end users.
## # It also allows for enabling options in one place.
## #
## FUBAR = system.getenv_bool("FUBAR", False,
##                            description="Fouled Up Beyond All Recognition processing")


class TestRgbColorName(TestWrapper):
    """Class for testcase definition"""
    script_file = TestWrapper.get_module_file_path(__file__)
    script_module = TestWrapper.get_testing_module_name(__file__)
    # TODO: use_temp_base_dir = True            # treat TEMP_BASE as directory
    # note: temp_file defined by parent (along with script_module, temp_base, and test_num)

    def test_data_file(self):
        """Makes sure TODO works as expected"""
        debug.trace(4, "TestRgbColorName.test_data_file()")

        content = (
            'Extracted colors:\n'
            '(255, 0, 0):  72.98% (1888)\n'
            '(0, 255, 0):  24.35% (630)\n'
            '(0, 0, 255):   2.67% (69)\n'
            '\n'
            'Pixels in output: 2587 of 11648\n'
        )
        system.write_file(self.temp_file, content)
        #   =>
        #   <(255, 0, 0), red>    :  33.33% (1)
        #   <(0, 255, 0), lime>    :  33.33% (1)
        #   <(0, 0, 255), blue>    :  33.33% (1)        
        output = self.run_script("", self.temp_file)
        assert re.search(
            r"<\(0, 255, 0\), lime>",
            output,
            )
        return

    @pytest.mark.skip(reason="TODO: test not yet implemented")
    def test_something_else(self):
        """TODO: flesh out test for something else"""
        debug.trace(4, "test_something_else()")
        assert(False, "TODO: code test")
        ## ex: assert THE_MODULE.TODO_function() == TODO_value
        return

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python
#
# Test(s) for ../misc_utils.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_misc_utils.py
#

"""Tests for misc_utils module"""

# Standard packages
## NOTE: this is empty for now

# Installed packages
import pytest

# Local packages
from mezcla import glue_helpers as gh
from mezcla import debug

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.misc_utils as THE_MODULE

class TestMiscUtils:
    """Class for test case definitions"""

    def test_transitive_closure(self):
        """Ensure transitive_closure works as expected"""
        debug.trace(4, "test_transitive_closure()")

        actual = THE_MODULE.transitive_closure([(1, 2), (2, 3), (3, 4)])
        expected = set([(1, 2), (1, 3), (1, 4), (2, 3), (3, 4), (2, 4)])
        assert actual == expected

    def test_read_tabular_data(self):
        """Ensure read_tabular_data works as expected"""
        debug.trace(4, "test_read_tabular_data()")
        string_table = (
            'language\tPython\n' +
            'framework\tPytest\n'
        )
        dict_table = {
            'language': 'Python\n',
            'framework': 'Pytest\n',
        }
        temp_file = gh.get_temp_file()
        gh.write_file(temp_file, string_table)
        assert THE_MODULE.read_tabular_data(temp_file) == dict_table

    def test_extract_string_list(self):
        """Ensure extract_string_list works as expected"""
        debug.trace(4, "test_extract_string_list()")
        assert THE_MODULE.extract_string_list("1  2,3") == ['1', '2', '3']

    def test_is_prime(self):
        """Ensure is_prime works as expected"""
        debug.trace(4, "test_is_prime()")

        FIRST_100_PRIMES = [
            2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37,
            41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83,
            89, 97, 101, 103, 107, 109, 113, 127, 131,
            137, 139, 149, 151, 157, 163, 167, 173, 179,
            181, 191, 193, 197, 199, 211, 223, 227, 229,
            233, 239, 241, 251, 257, 263, 269, 271, 277,
            281, 283, 293, 307, 311, 313, 317, 331, 337,
            347, 349, 353, 359, 367, 373, 379, 383, 389,
            397, 401, 409, 419, 421, 431, 433, 439, 443,
            449, 457, 461, 463, 467, 479, 487, 491, 499,
            503, 509, 521, 523, 541,
        ]

        assert all([THE_MODULE.is_prime(n) for n in FIRST_100_PRIMES])
        assert all([(not THE_MODULE.is_prime(n)) for n in range(FIRST_100_PRIMES[-1])  if n not in FIRST_100_PRIMES])

    def test_fibonacci(self):
        """Ensure fibonacci works as expected"""
        debug.trace(4, "test_fibonacci()")
        assert THE_MODULE.fibonacci(-5) == []
        assert THE_MODULE.fibonacci(1) == [0]
        assert THE_MODULE.fibonacci(10) == [0, 1, 1, 2, 3, 5, 8]

    def test_sort_weighted_hash(self):
        """Ensure sort_weighted_hash works as expected"""
        debug.trace(4, "test_sort_weighted_hash()")
        test_hash = {
            'bananas': 3,
            'apples': 1411,
            'peach': 43,
        }
        sorted_hash = [
            ('bananas', 3),
            ('peach', 43),
            ('apples', 1411),
        ]
        reversed_hash = [
            ('apples', 1411),
            ('peach', 43),
            ('bananas', 3),
        ]
        assert THE_MODULE.sort_weighted_hash(test_hash) == reversed_hash
        assert THE_MODULE.sort_weighted_hash(test_hash, reverse=False) == sorted_hash
        assert len(THE_MODULE.sort_weighted_hash(test_hash, max_num=2)) == 2

    def test_unzip(self):
        """Ensure unzip works as expected"""
        debug.trace(4, "test_unzip()")
        assert THE_MODULE.unzip(zip([1, 2, 3], ['a', 'b', 'c'])) == [[1, 2, 3], ['a', 'b', 'c']]
        assert THE_MODULE.unzip(zip([], []), 2) == [[], []]
        assert THE_MODULE.unzip(zip(), 4) == [[], [], [], []]
        assert THE_MODULE.unzip(zip(), 4) != [[], []]

    def test_get_current_frame(self):
        """Ensure get_current_frame works as expected"""
        debug.trace(4, "test_get_current_frame()")
        ## TODO: WORK-IN-PROGRESS

    def test_eval_expression(self):
        """Ensure eval_expression works as expected"""
        debug.trace(4, "test_eval_expression()")
        assert THE_MODULE.eval_expression("len([123, 321]) == 2")
        assert not THE_MODULE.eval_expression("'helloworld' == 2")

    def test_trace_named_object(self, capsys):
        """Ensure trace_named_object works as expected"""
        debug.trace(4, "test_trace_named_object()")
        # With level -1 we ensure that the trace will be printed
        THE_MODULE.trace_named_object(-1, "sys.argv")
        captured = capsys.readouterr()
        assert "sys.argv" in captured.err

    def test_trace_named_objects(self, capsys):
        """Ensure trace_named_objects works as expected"""
        debug.trace(4, "test_trace_named_objects()")
        # With level -1 we ensure that the trace will be printed
        THE_MODULE.trace_named_objects(-1, "[len(sys.argv), sys.argv]")
        captured = capsys.readouterr()
        assert "len(sys.argv)" in captured.err
        assert "sys.argv" in captured.err

    def test_exactly1(self):
        """Ensure exactly1 works as expected"""
        debug.trace(4, "test_exactly1()")
        assert THE_MODULE.exactly1([False, False, True])
        assert not THE_MODULE.exactly1([False, True, True])
        assert not THE_MODULE.exactly1([False, False, False])
        assert not THE_MODULE.exactly1([])

    def test_string_diff(self):
        """Ensure string_diff works as expected"""
        debug.trace(4, "test_string_diff()")

        STRING_ONE = 'one\ntwo\nthree\nfour'
        STRING_TWO = 'one\ntoo\ntree\nfour'
        EXPECTED_DIFF = (
            '  one\n'
            '< two\n'
            '…  ^\n'
            '> too\n'
            '…  ^\n'
            '< three\n'
            '…  -\n'
            '> tree\n'
            '  four\n'
        )

        assert THE_MODULE.string_diff(STRING_ONE, STRING_TWO) == EXPECTED_DIFF

    def test_elide_string_values(self):
        """Ensure elide_string_values works as expected"""
        debug.trace(4, "test_elide_string_values()")
        ## TODO: WORK-IN-PROGRESS


if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python
#
# Uses the Hugging Face API for automatic speech recognition (ASR).
#
# Based on following:
#   https://stackoverflow.com/questions/71568142/how-can-i-extract-and-store-the-text-generated-from-an-automatic-speech-recognit
#
# TODO:
# - Add chunking to handle large file:
#     https://huggingface.co/blog/asr-chunking
#

"""Speech recognition via Hugging Face"""

# Standard modules
# TODO: import re

# Intalled module
## OLD: import gradio as gr
## TODO:
## from transformers import pipeline

# Local modules
from mezcla import debug
from mezcla.main import Main
from mezcla import system
from mezcla import glue_helpers as gh

# Constants
TL = debug.TL

## TODO:
## # Environment options
## # Notes:
## # - These are just intended for internal options, not for end users.
## # - They also allow for enabling options in one place rather than four
## #   when using main.Main (e.g., [Main member] initialization, run-time
## #   value, and argument spec., along with string constant definition).
## #
## ENABLE_FUBAR = system.getenv_bool("ENABLE_FUBAR", False,
##                                   description="Enable fouled up beyond all recognition processing")

ASR_TASK = "automatic-speech-recognition"
# TODO: WHISPER = getenv...("whisper-large"); DEFAULT_MODEL = ...
DEFAULT_MODEL = "facebook/s2t-medium-librispeech-asr"
ASR_MODEL = system.getenv_text("ASR_MODEL", DEFAULT_MODEL,
                               "Hugging Face model for ASR")

#-------------------------------------------------------------------------------

SOUND_FILE = system.getenv_text("SOUND_FILE", "fuzzy-testing-1-2-3.wav",
                                "Audio file with speech to recognize")
USE_INTERFACE = system.getenv_bool("USE_INTERFACE", False,
                                   "Use web-based interface via gradio")

# Optionally load UI support
gr = None
if USE_INTERFACE:
    import gradio as gr                 # pylint: disable=import-error


def main():
    """Entry point"""
    debug.trace(TL.USUAL, f"main(): script={system.real_path(__file__)}")

    # Show simple usage if --help given
    dummy_app = Main(description=__doc__, skip_input=False, manual_input=False)

    # Resolve path for file
    sound_file = SOUND_FILE
    if not system.file_exists(sound_file):
        script_dir = gh.dirname(__file__)
        sound_file = gh.resolve_path(SOUND_FILE, base_dir=script_dir)
    if not system.file_exists(sound_file):
        system.exit(f"Error: unable to find SOUND_FILE '{sound_file}'")
    
    ## TEMP:
    ## pylint: disable=import-outside-toplevel
    from transformers import pipeline

    ## BAD:
    ## model = pipeline(task="automatic-speech-recognition",
    ##                  model="facebook/s2t-medium-librispeech-asr")
    model = pipeline(task=ASR_TASK, model=ASR_MODEL)

    if USE_INTERFACE:
        pipeline_if = gr.Interface.from_pipeline(
            model,
            title="Automatic Speech Recognition (ASR)",
            ## OLD: description="Using pipeline with Facebook S2T for ASR.",
            description="Using pipeline with default",
            examples=[sound_file])
        pipeline_if.launch()
    else:
        print(model(sound_file))

    return

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#! /usr/bin/env python
#
# Uses the Hugging Face API for machine translation (MT)
#
# Based on:
# - https://stackoverflow.com/questions/71568142/how-can-i-extract-and-store-the-text-generated-from-an-automatic-speech-recognit
# - Hugging Face's NLP with Transformers text
#

"""Machine translation via Hugging Face

Example:

echo "How now Bourne cow?" | FROM=en TO=es {script} -

USE_INTERFACE=1 {script} -
"""

# Standard modules
# TODO: import re

# Intalled module
## OLD: import gradio as gr
## TODO: import transformers
## OLD: from transformers import pipeline

# Local modules
from mezcla import debug
from mezcla.main import Main
from mezcla import misc_utils
from mezcla import system
from mezcla import glue_helpers as gh

# Constants
TL = debug.TL

## TODO:
## # Environment options
## # Notes:
## # - These are just intended for internal options, not for end users.
## # - They also allow for enabling options in one place rather than four
## #   when using main.Main (e.g., [Main member] initialization, run-time
## #   value, and argument spec., along with string constant definition).
## #
## ENABLE_FUBAR = system.getenv_bool("ENABLE_FUBAR", False,
##                                   description="Enable fouled up beyond all recognition processing")

FROM = system.getenv_text("FROM", "es")
TO = system.getenv_text("TO", "en")
SOURCE_LANG = system.getenv_text("SOURCE_LANG", FROM,
                                 "Source language")
TARGET_LANG = system.getenv_text("TARGET_LANG", TO,
                                 "Target language")
debug.assertion(SOURCE_LANG != TARGET_LANG)
MT_TASK = f"translation_{SOURCE_LANG}_to_{TARGET_LANG}"
DEFAULT_MODEL = f"Helsinki-NLP/opus-mt-{SOURCE_LANG}-{TARGET_LANG}"
MT_MODEL = system.getenv_text("MT_MODEL", DEFAULT_MODEL,
                              "Hugging Face model for MT")
SHOW_ELAPSED = system.getenv_bool("SHOW_ELAPSED", False,
                                  "Show elapsed time")
TEXT_ARG = "text"
## TODO: ELAPSED_ARG = "elapsed-time"

#-------------------------------------------------------------------------------

TEXT_FILE = system.getenv_text("TEXT_FILE", "-",
                               "Text file to translate")
USE_INTERFACE = system.getenv_bool("USE_INTERFACE", False,
                                   "Use web-based interface via gradio")

# Optionally load UI support
gr = None
if USE_INTERFACE:
    import gradio as gr                 # pylint: disable=import-error


def main():
    """Entry point"""
    debug.trace(TL.USUAL, f"main(): script={system.real_path(__file__)}")

    # Show simple usage if --help given
    ## OLD: dummy_app = Main(description=__doc__, skip_input=False, manual_input=False)
    dummy_app = Main(description=__doc__.format(script=__file__),
                     skip_input=False, manual_input=True,
                     ## TODO: bool_options=[(ELAPSED_ARG, "Show elapsed time")],
                     text_options=[(TEXT_ARG, "Text to translate")])
    debug.trace_object(5, dummy_app)
    debug.assertion(dummy_app.parsed_args)
    text = dummy_app.get_parsed_option(TEXT_ARG)

    # Get input file
    text_file = TEXT_FILE
    if ((text is not None) or USE_INTERFACE):
        pass
    elif (text_file == "-"):
        text_file = dummy_app.temp_file
        text = dummy_app.read_entire_input()
    else:
        text = system.read_file(text_file)

    ## TEMP:
    ## pylint: disable=import-outside-toplevel
    from transformers import pipeline
    model = pipeline(task=MT_TASK, model=MT_MODEL)

    if USE_INTERFACE:
        # TODO2: add language controls
        pipeline_if = gr.Interface.from_pipeline(
            model,
            title="Machine translation (MT)",
            ## TODO2: subtitle=f"From: {FROM}; To: {TO}",
            ## OLD:
            ## description="Using pipeline with default",
            description=f"From: {FROM}; To: {TO}",
            ## examples=[text_file])
            )
        pipeline_if.launch()
    else:
        TRANSLATION_TEXT = "translation_text"
        try:
            translation = model(text)
            debug.assertion(isinstance(translation, list)
                            and (TRANSLATION_TEXT in translation[0]))
            print(translation[0].get(TRANSLATION_TEXT) or "")
        except:
            system.print_exception_info("translation")
    debug.code(4, lambda: debug.trace(1, gh.run("nvidia-smi")))

    return

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    ## OLD: main()
    elapsed = misc_utils.time_function(main)
    if SHOW_ELAPSED:
       print(f"Elapsed time: {system.round_as_str(elapsed)}ms")



#! /usr/bin/env python
#
# Test(s) for ../pandas_sklearn.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_pandas_sklearn.py
#

"""Tests for pandas_sklearn module"""

# Standard packages
import re

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.unittest_wrapper import TestWrapper

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.pandas_sklearn as THE_MODULE

# Constants
EXAMPLES = f'{gh.dir_path(__file__)}/../examples'
RESOURCES = f'{gh.dir_path(__file__)}/resources'
IRIS_EXAMPLE = f'{EXAMPLES}/iris.csv'
IRIS_OUTPUT = f'{RESOURCES}/iris_output.txt'

class TestPandasSklearnUtils:
    """Class for testcase definition"""

    def test_create_feature_mapping(self):
        """Ensure create_feature_mapping works as expected"""
        debug.trace(4, "test_create_feature_mapping()")
        assert THE_MODULE.create_feature_mapping(['c', 'b', 'b', 'a']) == {'c':0, 'b':1, 'a':2}

    def test_show_ablation(self):
        """Ensure show_ablation works as expected"""
        debug.trace(4, "test_show_ablation()")
        ## TODO: WORK-IN-PROGRESS

    def test_show_precision_recall(self):
        """Ensure show_precision_recall works as expected"""
        debug.trace(4, "test_show_precision_recall()")
        ## TODO: WORK-IN-PROGRESS

    def test_show_average_precision_recall(self):
        """Ensure show_average_precision_recall works as expected"""
        debug.trace(4, "test_show_average_precision_recall()")
        ## TODO: WORK-IN-PROGRESS


class TestPandasSklearn(TestWrapper):
    """Class for testcase definition"""
    script_file = TestWrapper.get_module_file_path(__file__)
    script_module = TestWrapper.get_testing_module_name(__file__)

    def test_main_without_args(self):
        """Ensure main without args works as expected"""
        debug.trace(4, "test_main_without_args()")
        log_file = gh.get_temp_file()
        self.run_script(data_file='', log_file=log_file)
        assert 'Usage:' in gh.read_file(log_file)

    def test_normal_usage(self):
        """Ensure normal usage works as expected"""
        debug.trace(4, "test_normal_usage()")
        output = self.run_script(data_file=IRIS_EXAMPLE)
        assert output

        # Check line per line avoiding flaky numbers that always change
        for expected_line, actual_line in zip(gh.read_lines(IRIS_OUTPUT), output.splitlines()):
            expected_line = re.sub(r' +', ' ', expected_line)
            actual_line = re.sub(r' +', ' ', actual_line)
            if re.search(r'[0-9] +[0-9\.]+ +[0-9\.]+ +[0-9\.]+ +[0-9\.]+ +[a-zA-Z]+-[a-zA-Z]+', expected_line):
                continue
            assert actual_line == expected_line

    def test_verbose(self):
        """Ensure verbose works as expected"""
        debug.trace(4, "test_normal_usage()")
        output = self.run_script(env_options='VERBOSE=true', data_file=IRIS_EXAMPLE)
        assert 'Development test classification report:' in output
        expected_confusion_matrix = (
            'Development test set confusion matrix:\n'
            '[[3 0 0]\n'
            ' [0 5 0]\n'
            ' [0 1 3]]\n'
        )
        assert expected_confusion_matrix in output

    def test_show_ablation(self):
        """Ensure show_ablation works as expected"""
        debug.trace(4, "test_normal_usage()")
        expected_ablation_acc = (
            '[0.5, 0.43333333333333335, 0.5333333333333333,'
            ' 0.5, 0.5, 0.6, 0.6333333333333333, 0.6333333333333333,'
            ' 0.6333333333333333, 0.6333333333333333, 0.6333333333333333,'
            ' 0.6, 0.6, 0.6333333333333333, 0.6333333333333333,'
            ' 0.6333333333333333, 0.8666666666666667, 0.8333333333333334,'
            ' 0.9666666666666667, 0.9666666666666667, 0.9666666666666667,'
            ' 0.9666666666666667, 0.9333333333333333, 0.9333333333333333,'
            ' 0.9333333333333333, 0.9666666666666667, 0.9666666666666667,'
            ' 0.9666666666666667, 0.9333333333333333, 0.9333333333333333,'
            ' 0.9333333333333333, 0.9666666666666667, 0.9333333333333333,'
            ' 0.9333333333333333, 0.9333333333333333, 0.9333333333333333,'
            ' 0.9666666666666667, 0.9666666666666667, 0.9333333333333333,'
            ' 0.9333333333333333, 0.9333333333333333, 0.9333333333333333,'
            ' 0.9333333333333333, 0.9333333333333333, 0.9333333333333333,'
            ' 0.9333333333333333, 0.9333333333333333, 0.9333333333333333,'
            ' 0.9333333333333333, 0.9333333333333333, 0.9333333333333333,'
            ' 0.9333333333333333, 0.9333333333333333, 0.9333333333333333,'
            ' 0.9333333333333333, 0.9333333333333333, 0.9333333333333333,'
            ' 0.9333333333333333, 0.9333333333333333, 0.9333333333333333,'
            ' 0.9333333333333333, 0.9333333333333333, 0.9333333333333333,'
            ' 0.9333333333333333, 0.9333333333333333, 0.9333333333333333,'
            ' 0.9333333333333333, 0.9333333333333333, 0.9333333333333333,'
            ' 0.9333333333333333, 0.9333333333333333, 0.9333333333333333,'
            ' 0.9666666666666667, 0.9666666666666667, 1.0, 1.0, 1.0, 1.0,'
            ' 1.0, 0.9333333333333333, 0.9333333333333333, 0.9333333333333333,'
            ' 0.9333333333333333, 0.9333333333333333, 0.9666666666666667,'
            ' 0.9666666666666667, 0.9666666666666667, 1.0, 1.0, 1.0, 1.0,'
            ' 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,'
            ' 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]'
        )
        output = self.run_script(env_options='SHOW_ABLATION=true', data_file=IRIS_EXAMPLE)
        assert 'ablation accuracy:' in output 
        assert expected_ablation_acc in output

    def test_precision_recall(self):
        """Ensure precision_recall works as expected"""
        output = self.run_script(env_options='PRECISION_RECALL=true', data_file=IRIS_EXAMPLE)
        expected_content = (
            'precision:\n'
            '\t[]\n'
            'recall:\n'
            '\t[]\n'
            'thresholds:\n'
            '\t[]'
        )
        assert expected_content in output

    @pytest.mark.xfail
    def test_micro_average(self):
        """Ensure micro_average works as expected"""
        output = self.run_script(env_options='PRECISION_RECALL=true MICRO_AVERAGE=true', data_file=IRIS_EXAMPLE)
        expected_content = (
            'precision:\n'
            '\t[array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,\n'
            '       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1.])]\n'
            'recall:\n'
            '\t[array([1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.,\n'
            '       1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 0.])]\n'
            'thresholds:\n'
            '\t[array([7.13244554e-07, 1.87247554e-06, 3.51830227e-06, 4.18437007e-06,\n'
            '       6.44055641e-06, 8.99674077e-06, 1.03440695e-05, 1.33689339e-05,\n'
            '       5.65187581e-05, 8.06494481e-05, 6.66272967e-04, 1.19852249e-03,\n'
            '       1.93827217e-03, 2.30446715e-03, 5.68442555e-03, 6.21033670e-03,\n'
            '       1.15435434e-02, 1.86664414e-02, 2.75429206e-02, 1.36287126e-01,\n'
            '       9.43769992e-01, 9.43998991e-01, 9.50485768e-01, 9.63739944e-01,\n'
            '       9.66857292e-01, 9.71692022e-01, 9.72315717e-01, 9.72823159e-01,\n'
            '       9.73851932e-01, 9.84123197e-01])]\n'
            'average precision:\n'
            '\t[0.0]\n'
            'average recall:\n'
            '\t[1.0]'
        )
        # Replace very small numbers to avoid flaky results
        pattern = r'\d\d\d\d\de-'
        substitute = '00000e-'
        expected_content = re.sub(pattern, substitute, expected_content)
        output = re.sub(pattern, substitute, output)
        assert expected_content in output


if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python
#
# Runs training data through a variety of supervised classifiers, as an
# illustration on how to use pandas and sklearn to do machine learning.
#
# This was initially for the Iris dataset, based on following:
#    https://medium.com/codebagng/basic-analysis-of-the-iris-data-set-using-python-2995618a6342.
#
# Notes:
# - Environment variables:
#   DATA_FILE  FIELD_SEP  SCORING_METRIC  SEED  SKIP_DEVEL  SKIP_PLOTS  USE_DATAFRAME  VALIDATE_ALL  VALIDATION_CLASSIFIER  VERBOSE 
# - Currently only supports cross-validation (i.e., partitions of single datafile).
# - This partititions training data into development and validation sets.
# - Also does k-fold cross validation over development data split using 1/k-th as test.
# - That is, the training data is partitioned twice.
# - As an expediency to disable validation, epsilon is used for validation percent (e.g., 1e-6), because sklearn doesn't allow the percent to be specified as zero.
# - Keep changes in sync with pandas_sklearn.py (e.g., XGBoost and GPU options).
#................................................................................
# Debugging usage:
#
# module=pandas_sklearn
# let PSL++; dataset_path=examples/iris.csv; dataset_name=$(basename "$dataset_path" .csv); log="_$module.$dataset_name.$(TODAY).$PSL.log"; DEBUG_LEVEL=5 OUTPUT_CSV=1 XGB_SKIP_GPU=1 CLASSIFIER=XGB COERCE_FLOAT=1 ENCODE_CLASSES=1 VERBOSE=1 python -m mezcla.$module $dataset > "$log" 2>&1
#
#................................................................................
#
# TODO:
# - Add exception handling throughout.
# - Replace iris.csv with non-trivial ML data (e.g., to serve in benchmarking).
# - Make sure validation used in same tense as standard practice.
# - Rename (e.g., pandas_sklearn.py => evaluate_classifiers.py).
# - Decompose main() into more helpers.
# 

"""Illustrates sklearn classification over data with panda csv-based import"""

## TEST: allow for global logging.basicConfig initialization
from mezcla import debug            # pylint: disable=ungrouped-imports

# Standard packages
## OLD: import sys

# Installed pckages
import numpy as np
import pandas as pd
from pandas.plotting import scatter_matrix
from sklearn import model_selection
from sklearn.metrics import classification_report, confusion_matrix
from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score
from sklearn.metrics import precision_recall_curve
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier

from sklearn.neighbors import KNeighborsClassifier
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.naive_bayes import GaussianNB, MultinomialNB
from sklearn.svm import SVC

# Local packages
## OLD (put above for proper logging initialization):
## from mezcla import debug
from mezcla import data_utils as du
from mezcla import glue_helpers as gh
from mezcla import system
from mezcla import text_utils
from mezcla.system import round_num, getenv_bool, getenv_int, getenv_number, getenv_text, getenv_value
from mezcla.text_utils import getenv_ints

#................................................................................
# Constants (e.g., environment-based options)
#
# Note: Some classifiers are not part of sklearn and thus require additional support
# to enable. These include XBG, Keras, and Auto-Sklearn.

VERBOSE = getenv_bool("VERBOSE", False)
DATA_FILE = getenv_text("DATA_FILE", "examples/iris.csv")
FIELD_SEP = getenv_text("FIELD_SEP", ",",
                        description="Field separator (delimiter)")
SKIP_DATAFRAME = getenv_bool("SKIP_DATAFRAME", False)
USE_DATAFRAME = getenv_bool("USE_DATAFRAME", not SKIP_DATAFRAME)
IGNORE_FIELDS = text_utils.extract_string_list(
    getenv_value("IGNORE_FIELDS") or "")
INCLUDE_PLOTS = getenv_bool("INCLUDE_PLOTS", False)
DATASET_PLOTS = getenv_bool("DATASET_PLOTS", INCLUDE_PLOTS,
                            "Include plots summarizing dataset")
SKIP_PLOTS = getenv_bool("SKIP_PLOTS", not INCLUDE_PLOTS)
OUTPUT_CSV = getenv_bool("OUTPUT_CSV", False)
SKIP_VALIDATION = getenv_bool("SKIP_VALIDATION", False)
KERAS_CLASSIFIER = "Keras"
KNN_CLASSIFIER = "KNN"
XGB_CLASSIFIER = "XGB"
LR_CLASSIFIER = "LR"
DEFAULT_CLASSIFIER = getenv_text("CLASSIFIER", LR_CLASSIFIER,
                                 "Label for default classifier")
DEFAULT_DEVEL_CLASSIFIER = DEFAULT_CLASSIFIER if (not SKIP_VALIDATION) else None
DEVEL_CLASSIFIER = getenv_text("DEVEL_CLASSIFIER", DEFAULT_DEVEL_CLASSIFIER)
INCLUDE_ALL = getenv_bool("INCLUDE_ALL", False,
                          description="Include all classifiers (for devel and validation)")
INCLUDE_ALL_DEVEL = getenv_bool("INCLUDE_ALL_DEVEL", INCLUDE_ALL,
                                description="Evaluate all classifiers over development test data")
INCLUDE_MISC_CLASSIFIERS = getenv_bool("INCLUDE_MISC", INCLUDE_ALL,
                                       description="Include miscellaneous classifiers like CART")
EPSILON = 1.0e-6
DEFAULT_VALIDATION_PCT = 0.20 if (not SKIP_VALIDATION) else EPSILON
VALIDATION_PCT = getenv_number("VALIDATION_PCT", DEFAULT_VALIDATION_PCT)
VALIDATE_ALL = getenv_bool("VALIDATE_ALL", INCLUDE_ALL,
                           description="Evaluate all classifiers over validation over data")
DEFAULT_VALIDATION_CLASSIFIER = DEFAULT_CLASSIFIER if (not SKIP_VALIDATION) else None
VALIDATION_CLASSIFIER = getenv_text("VALIDATION_CLASSIFIER", DEFAULT_VALIDATION_CLASSIFIER)
## TEST: debug.assertion(False)
SKIP_DEVEL = getenv_bool("SKIP_DEVEL", False)
TEST_PCT = getenv_number("TEST_PCT", 0.10)
SEED = getenv_int("SEED", 7919)
SCORING_METRIC = getenv_text("SCORING_METRIC", "accuracy")
CLASS_VAR = getenv_value("CLASS_VAR", None,
                         "Classification variable name")
## TEST: 
## SKIP_XGB = getenv_bool("SKIP_XGB", False)
## USE_XGB = getenv_bool("USE_XGB", not SKIP_XGB)
DEFAULT_INCLUDE_XGB = XGB_CLASSIFIER in [DEFAULT_DEVEL_CLASSIFIER, DEFAULT_VALIDATION_CLASSIFIER]
INCLUDE_XGB = getenv_bool("INCLUDE_XGB", DEFAULT_INCLUDE_XGB or INCLUDE_ALL)
XGB_BOOSTER = getenv_value("XGB_BOOSTER", None)
XGB_SKIP_GPU = getenv_bool("XGB_SKIP_GPU", False)
XGB_USE_GPU = (not XGB_SKIP_GPU)
# TODO: standardize with respect to text_categorizer.py
XGB_USE_GPUS = XGB_USE_GPU
XGB_VERBOSITY = getenv_int("XGB_VERBOSITY", 0, "Degree of verbosity from 0 to 3")
#
COERCE_FLOAT = getenv_bool("COERCE_FLOAT", False,
                           "Coerce all data values to floating point, excluding class column")
ENCODE_CLASSES = getenv_bool("ENCODE_CLASSES", False,
                             "Encode classes using enumeration")
NUMERIC_CLASSES = getenv_bool("NUMERIC_CLASSES", False,
                              "Encode classes as number")
#
DEFAULT_INCLUDE_KERAS_CLASSIFIER = KERAS_CLASSIFIER in [DEFAULT_DEVEL_CLASSIFIER, DEFAULT_VALIDATION_CLASSIFIER]
INCLUDE_KERAS = getenv_bool("INCLUDE_KERAS", DEFAULT_INCLUDE_KERAS_CLASSIFIER or INCLUDE_ALL,
                            "Include Keras deep learning classifier")
SKIP_KERAS = getenv_bool("SKIP_KERAS", not INCLUDE_KERAS)
HIDDEN_UNIT_VALUES = getenv_ints("HIDDEN_UNIT_VALUES", "50, 100, 50")
NUM_EPOCHS = getenv_int("NUM_EPOCHS", 100)
BATCH_SIZE = getenv_int("BATCH_SIZE", None)
#
AUTOSKLEARN_CLASSIFIER = "Auto-Sklearn"
DEFAULT_INCLUDE_AUTOSKLEARN_CLASSIFIER = AUTOSKLEARN_CLASSIFIER in [DEFAULT_DEVEL_CLASSIFIER, DEFAULT_VALIDATION_CLASSIFIER]
INCLUDE_AUTOSKLEARN = getenv_bool("INCLUDE_AUTOSKLEARN", DEFAULT_INCLUDE_AUTOSKLEARN_CLASSIFIER,
                                  "Include Auto-Sklearn classifier for AutoML")
#
GPU_DEVICE = getenv_value("GPU_DEVICE", None,     # TODO: clarify value to use
                          "Device number for GPU (e.g., shown under nvidia-smi)")
SHOW_ABLATION = getenv_bool("SHOW_ABLATION", False,
                            "Show ablation plot for accuracy")
PRECISION_RECALL = getenv_bool("PRECISION_RECALL", False,
                               "Plot precision/recall curve")
MICRO_AVERAGE = getenv_bool("MICRO_AVERAGE", False,
                            "Use micro-averaging for mutliclass problems")
DUMP_MODEL = getenv_bool("DUMP_MODEL", False,
                         "Dump out model-specific representation")

# Globals
devel_classifiers = [DEVEL_CLASSIFIER]
validation_classifiers = [VALIDATION_CLASSIFIER]

#...............................................................................
# Optional packages

if INCLUDE_PLOTS:
    # pylint: disable=import-outside-toplevel, import-error
    import matplotlib.pyplot as plt

if INCLUDE_XGB:
    # pylint: disable=import-outside-toplevel, import-error
    import xgboost as xgb
    devel_classifiers.append(XGB_CLASSIFIER)
    validation_classifiers.append(XGB_CLASSIFIER)
if getenv_value("USE_XGB"):
    system.print_stderr("Warning: deprecated option USE_XGB")
## Mote: Following added for tracking down segmentation fault
## DEBUG: print("after xgboost import")
    
if INCLUDE_KERAS:
    # pylint: disable=import-outside-toplevel, ungrouped-imports, import-error
    from mezcla.keras_param_search import MyKerasClassifier, create_keras_model
    devel_classifiers.append(KERAS_CLASSIFIER)
    validation_classifiers.append(KERAS_CLASSIFIER)
## DEBUG: print("after keras_param_search import")

if INCLUDE_AUTOSKLEARN:
    # pylint: disable=import-outside-toplevel, import-error
    import autosklearn.classification
    devel_classifiers.append(AUTOSKLEARN_CLASSIFIER)
    validation_classifiers.append(AUTOSKLEARN_CLASSIFIER)

if (INCLUDE_ALL and (not (INCLUDE_XGB or INCLUDE_KERAS or INCLUDE_AUTOSKLEARN))):
    debug.trace(4, "Warning: Need to specify INCLUDE_(XGB/KERAS/AUTOSKLEARN) separately")

#...............................................................................
# Utility functions

def create_feature_mapping(label_values):
    """Return hash mapping elements from LABEL_VALUES into integers"""
    # EX: create_feature_mapping(['c', 'b, 'b', 'a']) => {'c':0, 'b':1, 'a':2}
    in_label_values = label_values
    if not isinstance(label_values, list):
        label_values = list(label_values)
    id_hash = {}
    for item in label_values:
        if (item not in id_hash):
            id_hash[item] = len(id_hash)
    debug.trace_fmtd(7, "create_feature_mapping({l}) => {h}", l=in_label_values, h=id_hash)
    return id_hash

#...............................................................................
# Main processing

def main():
    """Entry point for script"""
    debug.trace(4, "main()")

    # Misc. initialization
    system.PRECISION = 3

    # TODO: convert into using main.py's Main class
    data_file = None
    args = system.get_args()
    show_help = ((len(args) > 1) and (args[1] == "--help"))
    show_usage = ((len(args) > 1) and (args[1] == "--usage"))
    
    if ((len(args) <= 1) or show_help or show_usage):
        script = gh.basename(args[0])
        system.print_stderr("Usage: {scr} [--help | --usage] [data-file | -]", scr=script)
        system.print_stderr("")
        system.print_stderr("Notes:")
        system.print_stderr("- Use - for data-file to use default.")
        system.print_stderr("- Where data-file is in CSV format (or TSV)")
        ## TODO: show_details = VERBOSE or debug.debugging()
        debugging = debug.detailed_debugging()
        if (show_help or debugging):
            ## TODO: add legend
            system.print_stderr("- Environment options:")
            system.print_stderr("\t" + system.formatted_environment_option_descriptions(include_all=debugging))
        system.exit("")
    if ((len(args) > 1) and (not args[1].startswith("-"))):
        data_file = args[1]
    debug.trace(4, f"data_file={data_file}")
    if data_file is None:
        data_file = gh.resolve_path(DATA_FILE)

    # Read the data
    ##
    ## BAD
    ## extra_read_args = {}
    ## if COERCE_FLOAT:
    ##     extra_read_args['dtype'] = np.float64
    ## dataset = pandas.read_csv(DATA_FILE, sep=FIELD_SEP, comment="#", **extra_read_args)
    ##
    try:
        dataset = du.read_csv(data_file, sep=FIELD_SEP, comment="#")
    except:
        system.print_exception_info("du.read_csv")
        debug.trace(3, "Using pandas read_csv directly as fallback")
        dataset = pd.read_csv(data_file, sep=FIELD_SEP)

    # Reorder columns so classication variable last
    if (CLASS_VAR and (CLASS_VAR != dataset.columns[-1])):
        debug.trace_fmtd(4, "Re-arranging columns")
        # pylint: disable=no-member
        columns = system.difference(list(dataset.columns), CLASS_VAR)
        columns.append(CLASS_VAR)
        dataset = dataset[columns]

    # Show features proper and classificaiton variable
    feature_names = list(dataset.columns[0:-1])
    class_var = dataset.columns[-1]
    debug.trace_fmtd(4, "class_var={c} features:{f}",
                     f=feature_names, c=class_var)
    debug.trace_object(7, dataset, "dataset")

    # Optionally remove specified fields from data
    if IGNORE_FIELDS:
        debug.assertion(not system.difference(IGNORE_FIELDS, feature_names))
        dataset = dataset.drop(IGNORE_FIELDS, axis=1)
        feature_names = list(dataset.columns[0:-1])

    # Show samples from the data along with summary statistics and other information
    try:
        # Show first 10 rows
        print("Sample of data set (head, tail, random):")
        print(dataset.head())
        # SHow last 10 row
        print(dataset.tail())
        # Show 5 random rows
        print(dataset.sample(5))
        # Show a statistical summary about the dataset.
        print("statistical summary:")
        print(dataset.describe())
        # Show how many null entries are in the dataset.
        print("Null count:")
        print(dataset.isnull().sum())
    except:
        system.print_exception_info("dataset illustration")
    
    # Optionally show some plot
    ## OLD: if (not SKIP_PLOTS):
    if DATASET_PLOTS:
            
        # box and whisker plots
        dataset.plot(kind="box", subplots=True, layout=(2, 2), sharex=False, sharey=False)
        plt.show()
        
        # histograms
        dataset.hist()
        plt.show()
        
        # scatter plot matrix
        scatter_matrix(dataset)
        plt.show()

    # Split-out validation dataset
    if USE_DATAFRAME:
        features_indices = dataset.columns
        num_features = len(features_indices) - 1
        X = dataset[features_indices[0:num_features]]
        y = dataset[features_indices[num_features]]
        if COERCE_FLOAT:
            X = X.astype(np.float64)
        if NUMERIC_CLASSES:
            y = y.astype(np.float64)
    else:
        # pylint: disable=no-member
        array = dataset.values
        debug.trace_object(7, array, "array")
        num_features = (array.shape[1] - 1)
        X = array[:, 0:num_features]
        y = array[:, num_features]
        # TODO: COERCE_FLOAT
        # TODO: NUMERIC_CLASSES
    if ENCODE_CLASSES:
        y_encodings = create_feature_mapping(y)
        y_enum = sorted(y_encodings.keys(), key=lambda k: y_encodings[k])
        debug.trace_expr(4, y_encodings, y_enum)
        y = np.array([y_encodings[v] for v in y])
        if USE_DATAFRAME:
            y = pd.DataFrame(y, columns=[class_var])
        # TODO: COERCE_FLOAT
    debug.trace_fmtd(7, "X={X}\ny={y}", X=X, y=y)
    if USE_DATAFRAME:
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(y, columns=feature_names)
        if not isinstance(y, pd.DataFrame):
            y = pd.DataFrame(y, columns=[class_var])        
    if OUTPUT_CSV:
        # TODO: drop pandas index column (first one; no header)
        debug.assertion(USE_DATAFRAME)
        # note: prevously X and Y aren't frames if coerced or otherwise converted
        ## OLD:
        ## debug.assertion(not (COERCE_FLOAT or NUMERIC_CLASSES or ENCODE_CLASSES))
        debug.assertion((isinstance(X, pd.DataFrame) and isinstance(y, pd.DataFrame)))
        basename = system.remove_extension(data_file)
        # TODO: combine into single dataframe and use to_csv over that; use .csv instead of .csv.list (and make sure not to overwrite)
        X.to_csv(basename + "-X.csv.list", sep=FIELD_SEP, index=False)
        y.to_csv(basename + "-y.csv.list", sep=FIELD_SEP, index=False)
        gh.run("paste --delimiters='{d}' {b}-X.csv.list {b}-y.csv.list > {b}.csv.list",
               d=FIELD_SEP, b=basename)
        debug.assertion(system.file_exists(basename + ".csv.list"))
    X_train, X_validation, y_train, y_validation = model_selection.train_test_split(X, y, test_size=VALIDATION_PCT, random_state=SEED)
    ## TODO:
    debug.trace_fmtd(6, "X_train={xt}\nX_valid={xv}\ny_train={yt}\ny_valid={yv}", xt=X_train, xv=X_validation, yt=y_train, yv=y_validation)
    
    # Test options and evaluation metric
    ## NOTE: precision and recall currently not supported (see below)
    y_values = list(y.values) if isinstance(y, pd.Series) else y
    num_classes = len(create_feature_mapping(y_values))
    is_binary = (num_classes == 2)
    
    # Spot check algorithms
    # TODO: Add legend for non-standard abbreviations.
    models = []

    # TODO: drop n_gpus (thanks a lot, XGBoost: WTH?!)
    ## TEST: XGB_UPDATER = 'grow_gpu' if XGB_NUM_GPUS else ''
    ## TEST: models.append((XGB_CLASSIFIER, xgb.XGBClassifier(booster=XGB_BOOSTER, n_gpus=XGB_NUM_GPUS, updater=XGB_UPDATER)))
    if INCLUDE_XGB:
        ## OLD
        ## XGB_TREE_METHOD = 'gpu_hist' if XGB_USE_GPU else None
        ## models.append((XGB_CLASSIFIER, xgb.XGBClassifier(booster=XGB_BOOSTER, tree_method=XGB_TREE_METHOD)))
        ## TODO: misc_xgb_params = {}
        misc_xgb_params = {'booster': XGB_BOOSTER, 'verbosity': XGB_VERBOSITY}
        if XGB_USE_GPUS:
            misc_xgb_params.update({'tree_method': 'gpu_hist'})
            misc_xgb_params.update({'predictor': 'gpu_predictor'})
        if GPU_DEVICE:
            misc_xgb_params.update({'gpu_id': GPU_DEVICE})
        debug.trace_fmt(4, 'misc_xgb_params={m}', m=misc_xgb_params)
        models.append((XGB_CLASSIFIER, xgb.XGBClassifier(**misc_xgb_params)))
    if INCLUDE_KERAS:
        create_model_fn = lambda: create_keras_model(num_input_features=num_features,
                                                     num_classes=num_classes)
        models.append((KERAS_CLASSIFIER,
                       MyKerasClassifier(hidden_units=HIDDEN_UNIT_VALUES, epochs=NUM_EPOCHS,
                                         batch_size=BATCH_SIZE, build_fn=create_model_fn)))
    if (INCLUDE_ALL or (LR_CLASSIFIER in [DEFAULT_DEVEL_CLASSIFIER, DEFAULT_VALIDATION_CLASSIFIER])):
        models.append((LR_CLASSIFIER, LogisticRegression()))
    if INCLUDE_MISC_CLASSIFIERS:
        models.append(("LDA", LinearDiscriminantAnalysis()))
        models.append((KNN_CLASSIFIER, KNeighborsClassifier()))
        models.append(("CART", DecisionTreeClassifier()))        #  Classification and Regression Trees
        models.append(("GNB", GaussianNB()))
        models.append(("MNB", MultinomialNB()))
        models.append(("SVM", SVC()))
    if INCLUDE_AUTOSKLEARN:
        models.append((AUTOSKLEARN_CLASSIFIER, autosklearn.classification.AutoSklearnClassifier()))
    debug.trace_expr(5, models)
    if not models:
        system.exit("Error: no models defined")

    # Evaluate each model in turn.
    # TODO: show precision, recall, F1, as well as accuracy
    # Sample results:
    # name  acc    stdev
    # LR    0.967  0.041
    # LDA   0.975  0.038
    # KNN   0.983  0.033
    # CART  0.975  0.038
    # GNB   0.975  0.053
    # SVM   0.982  0.025
    #    
    summaries = []
    if (not SKIP_DEVEL):
        print("Sample development test set results using scoring method '{sm}'".format(sm=SCORING_METRIC))
        ## TODO: average = "micro" if (not is_binary) else None
        for name, model in models:
            if ((name not in devel_classifiers) and (not INCLUDE_ALL_DEVEL)):
                debug.trace_fmt(5, "Skipping classifier {n} (not for devel and not include all)", n=name)
                continue
            kfold = model_selection.KFold(n_splits=10, shuffle=True, random_state=SEED)
            ## TODO: cv_results = model_selection.cross_val_score(model, X_train, y_train, cv=kfold, scoring=SCORING_METRIC, average=average)
            ## TODO: get this to work when SCORING_METRIC is not accuracy (which leads to not supported error for multiclass data)
            ## (e.g., add environment variable so that sklearn uses micro or macro average
            try:
                cv_results = model_selection.cross_val_score(model, X_train, y_train, cv=kfold, scoring=SCORING_METRIC)
                summaries.append("{n}\t{avg}\t{std}".format(n=name, avg=system.round_num(cv_results.mean()), std=system.round_num(cv_results.std())))

                # Show confusion matrix for sample split of training data
                if VERBOSE:
                    X_devel, X_test, y_devel, y_test = model_selection.train_test_split(X_train, y_train, test_size=TEST_PCT, random_state=SEED)
                    model.fit(X_devel, y_devel)
                    debug.trace_fmtd(4, "devel data score: {s}", s=model.score(X_devel, y_devel))
                    model.fit(X_test, y_test)
                    debug.trace_fmtd(4, "test data score: {s}", s=model.score(X_test, y_test))
                    predictions = model.predict(X_test)
                    print("Development test set confusion matrix:")
                    print(confusion_matrix(y_test, predictions))
                    print("Development test classification report:")
                    print(classification_report(y_test, predictions))

                    # Model-specific information (e.g., feature importance)
                    if (name == XGB_CLASSIFIER):
                        print("Feature importance:")
                        ## print(model.get_score())
                        print(sorted(zip(feature_names, model.feature_importances_),
                                     key=lambda name_score: name_score[1],
                                     reverse=True))
                                          
            except:
                system.print_exception_info("training evaluation")
        print("Cross validation results over development test set")
        print("name\tacc\tstdev")
        print("\n".join(summaries))
    
    # Make predictions on validation dataset
    # Sample output:
    # Confusion matrix:
    # [[ 7  0 0]
    #  [ 0 11 1]
    #  [ 0  2 9]]
    # classification report:
    #              precision recall  f1  support
    # Iris-setosa     1.00   1.00  1.00  7
    # Iris-versicolor 0.85   0.92  0.88  12
    # Iris-virginica  0.90   0.82  0.86  11
    # avg / total     0.90   0.90  0.90  30
    # TODO: rework so that loop bypassed if no validation set
    average = "micro" if (not is_binary) else "binary"
    num_run = 0
    for name, model in models:
        if ((name not in validation_classifiers) and (not VALIDATE_ALL)):
            debug.trace_fmt(5, "Skipping classifier {n} (not for validation and not include all)", n=name)
            continue
        try:
            ## DEBUG: debug.trace(5, f"X/Y_train types: {[type(v) for v in [X_train, y_train]]}")
            ## DEBUG: debug.trace(5, f"X/Y_train head: {[v.head() for v in [X_train, y_train]]}")
            if VALIDATE_ALL:
                print("." * 80)
            print("Results over validation data for {n}:".format(n=name))
            num_run += 1
            model.fit(X_train, y_train)
            debug.trace_fmtd(4, "training data score: {s}", s=model.score(X_train, y_train))
            if debug.debugging(4):
                print("training confusion matrix:")
                training_predictions = model.predict(X_train)
                print(confusion_matrix(y_train, training_predictions))
            if DUMP_MODEL:
                if (name == XGB_CLASSIFIER):
                    basename = system.remove_extension(data_file)
                    model.get_booster().dump_model(f"{basename}.model-dump.list",
                                                   with_stats=True)
                else:
                    system.print_error(f"Warning: no model dump support for {name}")

            predictions = model.predict(X_validation)
            print("validation confusion matrix:")
            print(confusion_matrix(y_validation, predictions))
            ## TODO: drop accuracy ... F1 (provide in report)
            if VERBOSE:
                print("accuracy:", end=" ")
                print(round_num(accuracy_score(y_validation, predictions)))
                print("precision:", end=" ")
                print(round_num(precision_score(y_validation, predictions, average=average)))
                print("recall:", end=" ")
                print(round_num(recall_score(y_validation, predictions, average=average)))
                print("F1:", end=" ")
                print(round_num(f1_score(y_validation, predictions, average=average)))
            else:
                debug.trace_fmtd(4, "Use VERBOSE=1 for validation score breakdown")
            print("classification report:")
            print(classification_report(y_validation, predictions))
            if PRECISION_RECALL:
                if MICRO_AVERAGE:
                    show_average_precision_recall(name, model, num_classes, X_train, y_train, X_validation, y_validation)
                else:
                    show_precision_recall(name, model, num_classes, X_train, y_train, X_validation, y_validation)
            if SHOW_ABLATION:
                show_ablation(name, model, X_train, y_train, X_validation, y_validation)
        except:
            system.print_exception_info("validation evaluation")
    if ((num_run == 0) and VALIDATION_CLASSIFIER):
        system.print_stderr("Error: Validation classifier '{clf}' not supported", clf=VALIDATION_CLASSIFIER)

#------------------------------------------------------------------------

def show_ablation(name, model, X_train, y_train, X_validation, y_validation):
    """Show ablation result for X_TRAIN, Y_TRAIN, X_VALIDATION, Y_VALIDATION"""
    # Note: plots the data unless SKIP_PLOTS
    debug.trace(4, f"show_ablation{tuple([name, model, X_train, y_train, X_validation, y_validation])}")
    num_training = len(X_train)
    accuracies = []
    for size in range(num_training):
        try:
            model.fit(X_train[:size], y_train[:size])
            predictions = model.predict(X_validation)
            accuracy = accuracy_score(y_validation, predictions)
            accuracies.append(accuracy)
            debug.trace_expr(5, size, accuracy)
        except:
            system.print_exception_info("show_ablation")
    if (not SKIP_PLOTS):
        plt.plot(accuracies)
        plt.show()
    else:
        print("ablation accuracy:")
        print(accuracies)  
    debug.trace(5, "end show_ablation")
    return

def show_precision_recall(name, model, num_classes, X_train, y_train, X_validation, y_validation):
    """Show precision/recall results for X_TRAIN, Y_TRAIN, X_VALIDATION, Y_VALIDATION
    Note: This is only for binary classification
    """
    # based on https://www.statology.org/precision-recall-curve-python
    num_cases = len(y_validation)
    debug.assertion(num_classes == 2)
    precision = recall = thresholds = []
    debug.trace(4, f"show_precision_recall{tuple([name, model, num_classes, X_train, y_train, X_validation, y_validation])}")
    try:
        model.fit(X_train, y_train)
        y_scores = model.predict_proba(X_validation)[:, 1]
        precision, recall, thresholds = precision_recall_curve(y_validation, y_scores)
    except:
        system.print_exception_info("show_precision_recall")
    num_points = len(precision)
    debug.trace_expr(5, num_cases, num_points, len(precision), precision, recall, thresholds)
    ## TODO: debug.assertion(num_points <= (num_cases / 2))

    # Show the plots or print the data
    if (not SKIP_PLOTS):
        plt.plot(recall, precision)
        ## TODO: plt.yticks(thresholds)
        plt.xlabel("recall")
        plt.ylabel("precision")
        plt.title("precision recall curve")
        plt.show()
    else:
        ## TODO:
        ## print("precision, recall:")
        ## print(list(zip(precision, recall)))
        ## print(thresholds)
        print("precision:\n\t{vals!r}".format(vals=precision))
        print("recall:\n\t{vals!r}".format(vals=recall))
        print("thresholds:\n\t{vals!r}".format(vals=thresholds))
    debug.trace(5, "end show_precision_recall")        
    return
    

def show_average_precision_recall(name, model, num_classes, X_train, y_train, X_validation, y_validation):
    """Show precision/recall results for X_TRAIN, Y_TRAIN, X_VALIDATION, Y_VALIDATION
    Note: this is an approximation based on micro-averaging"""
    # based loosely on combination of https://www.statology.org/precision-recall-curve-python
    # and https://scikit-learn.org/stable/auto_examples/model_selection/plot_precision_recall.html
    debug.trace(4, f"show_average_precision_recall{tuple([name, model, num_classes, X_train, y_train, X_validation, y_validation])}")

    # Compute the metrics for each class
    num_cases = len(y_validation)
    precision = [None] * num_classes
    recall = [None] * num_classes
    thresholds = [None] * num_classes
    try:
        model.fit(X_train, y_train)
        for i in range(num_classes):
            y_scores = model.predict_proba(X_validation)[:, i]
            precision[i], recall[i], thresholds[i] = precision_recall_curve(y_validation, y_scores,
                                                                            pos_label=i)
    except:
        system.print_exception_info("show_average_precision_recall compute")
        
    # Compute micro-average, quantifying score on all classes jointly
    # TODO: do summation on previous loop
    num_points = len(precision)
    average_precision = [0] * num_points
    average_recall = [0] * num_points
    try:
        for s in range(num_points):
            for i in range(num_classes):
                average_precision[s] += precision[i][s]
                average_recall[s] += recall[i][s]
            average_precision[s] /= num_classes
            average_recall[s] /= num_classes
    except:
        system.print_exception_info("show_average_precision_recall micro")
    debug.trace_expr(5, num_cases, num_points, len(precision), precision, recall, thresholds,
                     average_precision, average_recall)
    ## TODO: debug.assertion(num_points <= (num_cases / 2))

    # Show the plots or print the data
    if (not SKIP_PLOTS):
        try:
            plt.plot(average_recall, average_precision)
            plt.xlabel("mean recall")
            plt.ylabel("mean precision")
            plt.title("precision recall curve [micro-averaged]")
            plt.show()
        except:
            system.print_exception_info("show_average_precision_recall plot")
    else:
        print("precision:\n\t{vals!r}".format(vals=precision))
        print("recall:\n\t{vals!r}".format(vals=recall))
        print("thresholds:\n\t{vals!r}".format(vals=thresholds))
        print("average precision:\n\t{vals!r}".format(vals=average_precision))
        print("average recall:\n\t{vals!r}".format(vals=average_recall))
    debug.trace(5, "end show_average_precision_recall")        
    return

#------------------------------------------------------------------------

if __name__ == "__main__":
    main()



#! /usr/bin/env python
#
# Test(s) for ../__init__.py
#
# Notes:
# - Fill out TODO's below. Use numbered tests to order (e.g., test_1_usage).
# - TODO: If any of the setup/cleanup methods defined, make sure to invoke base
#   (see examples below for setUp and tearDown).
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test___init__.py
#

"""Tests for __init__ module"""

# Standard packages
import re

# Installed packages
import pytest

# Local packages
from mezcla.unittest_wrapper import TestWrapper
from mezcla import debug
## TODO: from mezcla import system
from mezcla.my_regex import my_re

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                  global module object
#    TestTemplate.script_module:  path to file
import mezcla.__init__ as THE_MODULE
#
# Note: sanity test for customization (TODO: remove if desired)
if not re.search(__file__, r"\btemplate.py$"):
    debug.assertion("mezcla.template" not in str(THE_MODULE))

## TODO:
## # Environment options
## # Note: These are just intended for internal options, not for end users.
## # It also allows for enabling options in one place.
## #
## FUBAR = system.getenv_bool("FUBAR", False,
##                            description="Fouled Up Beyond All Recognition processing")


class TestIt(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__)
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_data_file(self):
        """Makes sure TODO works as expected"""
        debug.trace(4, f"TestIt.test_data_file(); self={self}")
        output = self.run_script("", self.temp_file)
        assert(not output.strip())
        return

class TestIt2:
    """Another class for testcase definition
    Note: Needed to avoid error with pytest due to inheritance with unittest.TestCase via TestWrapper"""

    def test_version(self):
        """Test version string"""
        debug.trace(5, f"test_version(); self={self}")
        assert(hasattr(THE_MODULE, "__VERSION__"))
        assert(my_re.match(r"^\d+.\d+\.\d+.*", THE_MODULE.__VERSION__))


#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])
